@@ -35,15 +35,17 @@ Example Usage:
     result = await client.execute_task("Create a web scraper", continuous=False)
     ```
 """
+# Where's the workspace path?
 
 import asyncio
 import logging
 from pathlib import Path
-from typing import Any, Dict, List, Optional, Union, Callable, AsyncGenerator
+from typing import Any, Dict, List, Optional, Sequence, Union, Callable, AsyncGenerator
 from dataclasses import dataclass
 
 from .core import PenguinCore
 from .config import config
+from .llm.model_config import ModelConfig
 from .system.checkpoint_manager import CheckpointType
 
 logger = logging.getLogger(__name__)
@@ -58,6 +60,7 @@ class ChatOptions:
     streaming: bool = False
     max_iterations: int = 5
     image_path: Optional[str] = None
+    agent_id: Optional[str] = None
 
 
 @dataclass  
@@ -93,7 +96,7 @@ class ModelInfo:
 
 
 class PenguinClient:
-    """High-level Python client for Penguin AI functionality."""
+    """High-level Python client for Penguin functionality."""
     
     def __init__(
         self,
@@ -168,6 +171,7 @@ class PenguinClient:
             message=message,
             context=opts.context,
             conversation_id=opts.conversation_id,
+            agent_id=opts.agent_id,
             context_files=opts.context_files,
             streaming=opts.streaming
         )
@@ -202,6 +206,7 @@ class PenguinClient:
                 input_data={"text": message, **({"image_path": opts.image_path} if opts.image_path else {})},
                 context=opts.context,
                 conversation_id=opts.conversation_id,
+                agent_id=opts.agent_id,
                 max_iterations=opts.max_iterations,
                 context_files=opts.context_files,
                 streaming=True,
@@ -478,6 +483,184 @@ class PenguinClient:
     async def list_context_files(self) -> List[str]:
         """List available context files."""
         return self.core.list_context_files()
+
+    # Agent management helpers
+    # ------------------------------------------------------------------
+
+    def create_agent(
+        self,
+        agent_id: str,
+        *,
+        system_prompt: Optional[str] = None,
+        activate: bool = False,
+        share_session_with: Optional[str] = None,
+        share_context_window_with: Optional[str] = None,
+        shared_cw_max_tokens: Optional[int] = None,
+        model_max_tokens: Optional[int] = None,
+        persona: Optional[str] = None,
+        model_config: Optional[ModelConfig] = None,
+        model_config_id: Optional[str] = None,
+        model_overrides: Optional[Dict[str, Any]] = None,
+        default_tools: Optional[Sequence[str]] = None,
+    ) -> None:
+        """Register a new agent."""
+        self.core.register_agent(
+            agent_id,
+            system_prompt=system_prompt,
+            activate=activate,
+            share_session_with=share_session_with,
+            share_context_window_with=share_context_window_with,
+            shared_cw_max_tokens=shared_cw_max_tokens,
+            model_max_tokens=model_max_tokens,
+            persona=persona,
+            model_config=model_config,
+            model_config_id=model_config_id,
+            model_overrides=model_overrides,
+            default_tools=default_tools,
+        )
+
+    def create_sub_agent(
+        self,
+        agent_id: str,
+        *,
+        parent_agent_id: str,
+        system_prompt: Optional[str] = None,
+        share_session: bool = True,
+        share_context_window: bool = True,
+        shared_cw_max_tokens: Optional[int] = None,
+        model_max_tokens: Optional[int] = None,
+        activate: bool = False,
+        persona: Optional[str] = None,
+        model_config: Optional[ModelConfig] = None,
+        model_config_id: Optional[str] = None,
+        model_overrides: Optional[Dict[str, Any]] = None,
+        default_tools: Optional[Sequence[str]] = None,
+    ) -> None:
+        """Register a sub-agent bound to a parent agent."""
+        self.core.create_sub_agent(
+            agent_id,
+            parent_agent_id=parent_agent_id,
+            system_prompt=system_prompt,
+            share_session=share_session,
+            share_context_window=share_context_window,
+            shared_cw_max_tokens=shared_cw_max_tokens,
+            model_max_tokens=model_max_tokens,
+            activate=activate,
+            persona=persona,
+            model_config=model_config,
+            model_config_id=model_config_id,
+            model_overrides=model_overrides,
+            default_tools=default_tools,
+        )
+
+    def list_agents(self) -> List[str]:
+        """List registered agent identifiers."""
+        return self.core.list_agents()
+
+    def list_sub_agents(self, parent_agent_id: Optional[str] = None) -> Dict[str, List[str]]:
+        """List sub-agents mapped to their parent agents."""
+        return self.core.list_sub_agents(parent_agent_id)
+
+    # Agent roster and profiles
+    # ------------------------------------------------------------------
+
+    def get_agent_roster(self) -> List[Dict[str, Any]]:
+        """Return detailed roster for all registered agents (JSON friendly)."""
+        return self.core.get_agent_roster()
+
+    def get_agent_profile(self, agent_id: str) -> Optional[Dict[str, Any]]:
+        """Return roster information for a single agent."""
+        return self.core.get_agent_profile(agent_id)
+
+    # Agent pause/resume helpers
+    # ------------------------------------------------------------------
+
+    def pause_agent(self, agent_id: str) -> None:
+        """Mark an agent as paused (engine-driven actions should be skipped)."""
+        self.core.set_agent_paused(agent_id, True)
+
+    def resume_agent(self, agent_id: str) -> None:
+        """Mark an agent as resumed."""
+        self.core.set_agent_paused(agent_id, False)
+
+    async def unregister_agent(self, agent_id: str, *, preserve_conversation: bool = False) -> bool:
+        """Unregister an agent and optionally remove its conversation state."""
+        result = self.core.unregister_agent(agent_id, preserve_conversation=preserve_conversation)
+        if asyncio.iscoroutine(result):
+            result = await result
+        return bool(result)
+
+    async def send_to_agent(
+        self,
+        agent_id: str,
+        content: Any,
+        *,
+        message_type: str = "message",
+        metadata: Optional[Dict[str, Any]] = None,
+        channel: Optional[str] = None,
+    ) -> bool:
+        """Forward a structured message to an agent via MessageBus."""
+        return await self.core.send_to_agent(
+            agent_id,
+            content,
+            message_type=message_type,
+            metadata=metadata,
+            channel=channel,
+        )
+
+    async def send_to_human(
+        self,
+        content: Any,
+        *,
+        message_type: str = "status",
+        metadata: Optional[Dict[str, Any]] = None,
+        channel: Optional[str] = None,
+    ) -> bool:
+        """Forward a structured message to the human recipient via MessageBus."""
+        return await self.core.send_to_human(
+            content,
+            message_type=message_type,
+            metadata=metadata,
+            channel=channel,
+        )
+
+    # What if there's multiple humans? @penguin_todo_multi_humans.md 
+    # I suppose we could just have a human_id and then we can have multiple humans that way for now.
+    async def human_reply(
+        self,
+        agent_id: str,
+        content: Any,
+        *,
+        message_type: str = "message",
+        metadata: Optional[Dict[str, Any]] = None,
+        channel: Optional[str] = None,
+    ) -> bool:
+        """Send a human-authored reply to an agent via MessageBus."""
+        return await self.core.human_reply(
+            agent_id,
+            content,
+            message_type=message_type,
+            metadata=metadata,
+            channel=channel,
+        )
+
+    async def get_telemetry_summary(self) -> Dict[str, Any]:
+        """Fetch aggregated telemetry summary."""
+        return await self.core.get_telemetry_summary()
+
+    def get_conversation_history(
+        self,
+        conversation_id: str,
+        *,
+        include_system: bool = True,
+        limit: Optional[int] = None,
+    ) -> List[Dict[str, Any]]:
+        """Retrieve detailed conversation history annotated with agent metadata."""
+        return self.core.get_conversation_history(
+            conversation_id,
+            include_system=include_system,
+            limit=limit,
+        )
     
     # Utility Methods
     # ------------------------------------------------------------------
@@ -521,4 +704,4 @@ async def create_client(
         workspace_path=workspace_path
     )
     await client.initialize()
-    return client
\ No newline at end of file
+    return client
