@@ -19,6 +19,9 @@ import rich.box # Added for table box style
 from rich.live import Live # type: ignore
 from rich.layout import Layout
 
+# Import unified renderer
+from penguin.cli.renderer import UnifiedRenderer, RenderStyle
+
 # Add project root to sys.path - This might not be needed if ui.py is part of the package
 # and imported correctly by the main CLI. For now, let's assume direct imports work
 # or the main CLI handles path adjustments.
@@ -113,12 +116,24 @@ ROLE_THEME_MAP = {
 
 class CLIRenderer:
     """Handles all Rich-based rendering for the CLI"""
-    
+
+    def _load_cli_display_config(self) -> Dict[str, Any]:
+        """Load CLI display configuration from config file."""
+        try:
+            from penguin.config import config
+            cli_config = config.get('cli', {}) if isinstance(config, dict) else {}
+            display_config = cli_config.get('display', {})
+            logger.debug(f"Loaded CLI display config: {display_config}")
+            return display_config
+        except Exception as e:
+            logger.warning(f"Failed to load CLI display config: {e}, using defaults")
+            return {}
+
     def __init__(self, console: Console, core: "PenguinCore"):
         """
         Initialize the renderer with a console object and a core instance.
         Subscribes to Core events for UI updates.
-        
+
         Args:
             console: Rich console for rendering
             core: PenguinCore instance to subscribe to
@@ -126,10 +141,30 @@ class CLIRenderer:
         logger.debug("Initializing CLIRenderer")
         self.console = console
         self.core = core
-        
+
+        # Load CLI display settings from config
+        cli_config = self._load_cli_display_config()
+
+        # Initialize unified renderer with settings from config
+        self.renderer = UnifiedRenderer(
+            console=self.console,
+            style=RenderStyle[cli_config.get('style', 'MINIMAL').upper()],
+            show_timestamps=cli_config.get('show_timestamps', True),
+            show_metadata=cli_config.get('show_metadata', False),
+            show_tool_results=getattr(core, "show_tool_results", True),
+            filter_internal_markers=cli_config.get('hide_internal_markers', True),
+            deduplicate_messages=cli_config.get('deduplicate_messages', True),
+            max_blank_lines=cli_config.get('max_blank_lines', 2)
+        )
+
         # Streaming state - simplified
         self.streaming_message_data = {}
         self.is_streaming = False
+
+        # Event buffering for system messages (to consolidate consecutive system outputs)
+        self._pending_system_events: List[Dict[str, Any]] = []
+        self._system_event_timer: Optional[asyncio.Task] = None
+        self._last_message_role: Optional[str] = None
         
         # UI display controls
         self.show_context_messages: bool = False
@@ -169,11 +204,19 @@ class CLIRenderer:
         # Messages cache from conversation manager
         self.conversation_messages: List[Dict[str, Any]] = []
         
-        # Register with Core for events
+        # Register with unified event bus for all events
+        from penguin.cli.events import EventBus, EventType
+        event_bus = EventBus.get_sync()
+
+        # Subscribe to relevant event types
+        for event_type in EventType:
+            event_bus.subscribe(event_type.value, self.handle_event)
+        logger.debug("Subscribed to unified event bus for all UI events")
+
+        # Also register with Core for backward compatibility
         if self.core:
-            # Register the synchronous wrapper instead of the async method
-            self.core.register_ui(self.handle_event_sync)
-            logger.debug("Registered with Core for UI events")
+            self.core.register_ui(self.handle_event)
+            logger.debug("Also registered with Core for backward compatibility")
         
         # Call update once to build initial state
         self.update_token_stats(None)
@@ -200,6 +243,10 @@ class CLIRenderer:
                 await self._handle_status_event(data)
             elif event_type == "error":
                 await self._handle_error_event(data)
+            elif event_type == "tool":
+                # Tool events are handled by the Ink CLI EventTimeline component
+                # They're passed through to the frontend, so we just acknowledge them here
+                logger.debug(f"Tool event received: {data.get('action', 'unknown')}")
             else:
                 logger.warning(f"Unknown event type: {event_type}")
             
@@ -279,10 +326,73 @@ class CLIRenderer:
             logger.error(f"Error handling token event: {e}", exc_info=True)
 
     async def _handle_message_event(self, data: Dict[str, Any]) -> None:
-        """Handle new message events."""
-        # Messages are added to conversation manager by Core
-        # Just refresh our local cache
-        self._refresh_message_cache()
+        """Handle new message events with buffering for system messages."""
+        message_role = data.get("role", "")
+        category = data.get("category", "")
+
+        # Check if this is a system output message that should be buffered
+        if self._should_buffer_message(message_role, category):
+            await self._buffer_system_event(data)
+        else:
+            # Flush any pending system events first
+            await self._flush_system_events()
+            # Then refresh cache for this message
+            self._refresh_message_cache()
+            # Track this message role for separator logic
+            self._last_message_role = message_role
+
+    async def _buffer_system_event(self, data: Dict[str, Any]) -> None:
+        """
+        Buffer a system message event to potentially consolidate with others.
+
+        Args:
+            data: Message event data
+        """
+        self._pending_system_events.append(data)
+
+        # Cancel existing timer if any
+        if self._system_event_timer and not self._system_event_timer.done():
+            self._system_event_timer.cancel()
+
+        # Set new timer to flush after 100ms of no new system messages
+        self._system_event_timer = asyncio.create_task(self._delayed_flush_system_events())
+
+    async def _delayed_flush_system_events(self) -> None:
+        """Wait 100ms then flush pending system events."""
+        try:
+            await asyncio.sleep(0.1)  # 100ms delay
+            await self._flush_system_events()
+        except asyncio.CancelledError:
+            # Timer was cancelled, which is fine
+            pass
+
+    async def _flush_system_events(self) -> None:
+        """Flush all pending system events and refresh message cache."""
+        if self._pending_system_events:
+            logger.debug(f"Flushing {len(self._pending_system_events)} buffered system events")
+            self._pending_system_events.clear()
+            self._refresh_message_cache()
+
+    def _should_buffer_message(self, role: str, category: Any) -> bool:
+        """
+        Determine if a message should be buffered.
+
+        Args:
+            role: Message role
+            category: Message category
+
+        Returns:
+            True if message should be buffered
+        """
+        # Buffer consecutive system output messages
+        if role == "system":
+            # Check category (handle both enum and string forms)
+            if isinstance(category, MessageCategory):
+                return category == MessageCategory.SYSTEM_OUTPUT
+            elif isinstance(category, str):
+                return category == "SYSTEM_OUTPUT"
+
+        return False
 
     async def _handle_status_event(self, data: Dict[str, Any]) -> None:
         """Handle status update events."""
@@ -371,71 +481,12 @@ class CLIRenderer:
         return layout
 
     def _render_text_segment(self, text_content: str) -> List[Any]:
-        # This new helper contains the original logic for processing a string
-        # (finding code blocks, rendering markdown for parts)
-        renderables = []
-        last_end = 0
-
-        for match in CODE_BLOCK_PATTERNS.finditer(text_content):
-            start, end = match.span()
-            lang = match.group(1)
-            code = match.group(2)
-
-            if start > last_end:
-                preceding_text = text_content[last_end:start].strip()
-                if preceding_text:
-                    renderables.append(Markdown(preceding_text))
-            
-            display_lang = LANGUAGE_DISPLAY_NAMES.get(lang, lang.capitalize() if lang else "Code")
-            syntax = Syntax(code, lexer=lang if lang else "text", theme="default", line_numbers=True, word_wrap=True)
-            code_panel = Panel(syntax, title=display_lang, border_style=THEME_COLORS.get("code_border", "dim blue"), padding=(0, 1), expand=False)
-            renderables.append(code_panel)
-            last_end = end
-
-        if last_end < len(text_content):
-            remaining_text = text_content[last_end:].strip()
-            if remaining_text:
-                renderables.append(Markdown(remaining_text))
-
-        if not renderables and text_content: # Whole thing is markdown (no code blocks)
-             renderables.append(Markdown(text_content.strip()))
-        # Removed the empty text segment case as _render_message_content handles overall empty/waiting states
-
-        return renderables
+        """Use unified renderer for text segment rendering"""
+        return self.renderer._render_text_segment(text_content)
 
     def _render_message_content(self, content: Any, role_for_theme: str = "default") -> Group:
-        """Renders message content, handling Markdown, code blocks, and multimodal list content."""
-        overall_renderables = []
-
-        if isinstance(content, list): # Multimodal content
-            for item in content:
-                item_type = item.get("type")
-                if item_type == "text":
-                    text_segment = item.get("text", "")
-                    if text_segment: 
-                        segment_renderables = self._render_text_segment(text_segment)
-                        overall_renderables.extend(segment_renderables)
-                elif item_type == "image_url":
-                    image_path = item.get("image_path", "unknown image")
-                    overall_renderables.append(Text(f"[Image: {image_path}]", style="dim italic"))
-                # else: handle other future multimodal types like audio, video etc.
-            if not overall_renderables: 
-                overall_renderables.append(Text("(Empty or unrenderable multimodal content)", style="dim italic"))
-        elif isinstance(content, str): # Existing string content
-            string_renderables = self._render_text_segment(content)
-            overall_renderables.extend(string_renderables)
-            if not overall_renderables and not content: # String content was empty
-                 overall_renderables.append(Text("(Waiting for response...)", style="dim italic"))
-            elif not overall_renderables and content: # String content not empty but _render_text_segment returned nothing (e.g. just whitespace)
-                 overall_renderables.append(Markdown(content.strip())) # Try rendering the stripped original as Markdown
-
-        else: # Fallback for unexpected content type or if content is None/empty after processing
-            if not content and not overall_renderables: # Specifically if initial content was None or empty string
-                overall_renderables.append(Text("(Waiting for response...)", style="dim italic"))
-            elif not overall_renderables: # Content was of an unsupported type, or other empty case
-                overall_renderables.append(Text(f"(Unsupported or empty content: type {type(content)})", style="dim italic error"))
-            
-        return Group(*overall_renderables)
+        """Use unified renderer for message content"""
+        return self.renderer.render_content(content, role_for_theme)
 
     def _build_conversation_area_content(self) -> Group:
         """
@@ -445,27 +496,24 @@ class CLIRenderer:
         logger.debug("Building conversation area content")
         
         rendered_panels = []
-        
+
         # Filter system/context messages if needed
-        # IMPORTANT: System output (tool results) should always be shown!
+        # MINIMAL MODE: Hide ALL system output (tool results) for clean output
         filtered_messages = self.conversation_messages
         if not self.show_context_messages:
             filtered_messages = [
                 msg for msg in filtered_messages
-                if (isinstance(msg.get("category"), MessageCategory) and 
-                    msg.get("category") not in [MessageCategory.SYSTEM, MessageCategory.CONTEXT]) or
-                   (isinstance(msg.get("category"), str) and 
-                    msg.get("category") not in ["SYSTEM", "CONTEXT"]) or
-                   # Always show system output messages (tool results)
-                   (isinstance(msg.get("category"), MessageCategory) and 
-                    msg.get("category") == MessageCategory.SYSTEM_OUTPUT) or
-                   (isinstance(msg.get("category"), str) and 
-                    msg.get("category") == "SYSTEM_OUTPUT")
+                if (isinstance(msg.get("category"), MessageCategory) and
+                    msg.get("category") not in [MessageCategory.SYSTEM, MessageCategory.CONTEXT, MessageCategory.SYSTEM_OUTPUT]) or
+                   (isinstance(msg.get("category"), str) and
+                    msg.get("category") not in ["SYSTEM", "CONTEXT", "SYSTEM_OUTPUT"])
             ]
         
-        # Render each message as a panel
+        # Render each message as a panel (skip duplicates)
         for msg in filtered_messages:
-            rendered_panels.append(self.render_message_panel(msg))
+            rendered = self.render_message_panel(msg)
+            if rendered is not None:  # Skip duplicates
+                rendered_panels.append(rendered)
         
         # Add streaming message if active
         if self.is_streaming and self.streaming_message_data:
@@ -624,65 +672,27 @@ class CLIRenderer:
         """Returns the main renderable for the Live display."""
         return self.get_layout_renderable()
 
-    def render_message_panel(self, message_data: Dict[str, Any]) -> Panel:
+    def render_message_panel(self, message_data: Dict[str, Any]) -> Union[Panel, Group, None]:
         """
-        Renders a single message into a Rich Panel.
-        Handles different roles and formats content with Markdown/code blocks.
+        Use unified renderer for message panels with deduplication.
+
+        Args:
+            message_data: Message data dict
+
+        Returns:
+            Rendered panel, group, or None if duplicate
         """
-        logger.debug(f"Rendering message panel for role: {message_data.get('role', 'unknown')}")
-        
-        # Extract message data
-        role = message_data.get("role", "unknown").lower()
-        content = message_data.get("content", "")
-        category_name = message_data.get("category", MessageCategory.UNKNOWN.name) # Get category name
-        if isinstance(category_name, MessageCategory):
-            category_name = category_name.name
-        timestamp_str = message_data.get("timestamp", "")
-        metadata = message_data.get("metadata", {})
-
-        # Determine panel title and style based on role and category
-        panel_title_text = role.capitalize()
-        if role == "system" and category_name:
-            panel_title_text = f"System ({category_name.replace('_', ' ').title()})"
-        elif role == "assistant":
-            panel_title_text = "Penguin"
-            if metadata.get("tool_name"):
-                panel_title_text = f"Penguin (Tool: {metadata.get('tool_name')})"
-            elif metadata.get("is_streaming"):
-                panel_title_text = "Penguin (streaming)"
-
-        panel_border_style = THEME_COLORS.get(f"message_panel_{role}", THEME_COLORS.get("message_panel_default", "white"))
-        if role == "system" and category_name == MessageCategory.ERROR.name:
-            panel_border_style = THEME_COLORS.get("error", "bold red")
-        
-        # Special handling for streaming assistant message
-        is_streaming = metadata.get("is_streaming", False)
-        text_to_render = content
-        if is_streaming and role == "assistant":
-            if not text_to_render.strip():
-                text_to_render = "Waiting for response... 🐧"
-            elif not text_to_render.endswith("🐧"):
-                text_to_render += " 🐧"
-
-        rendered_content_group = self._render_message_content(text_to_render, role_for_theme=role)
-
-        # Add timestamp to title if available
-        if timestamp_str:
-            formatted_time = self._safe_format_timestamp(timestamp_str)
-            if formatted_time:
-                panel_title_text += f" @ {formatted_time}"
-
-        # Add logging for panels
-        logger.debug(f"Creating panel with title: {panel_title_text}, role: {role}, category: {category_name}")
-
-        return Panel(
-            rendered_content_group,
-            title=panel_title_text,
-            border_style=panel_border_style,
-            padding=(0,1), # Minimal vertical padding
-            expand=True # Allow panel to expand horizontally
+        rendered = self.renderer.render_message(
+            message_data,
+            as_panel=True
         )
 
+        # Update last message role for separator logic
+        if rendered is not None:
+            self._last_message_role = message_data.get("role")
+
+        return rendered
+
     def render_message_list(self, messages: List[Dict[str, Any]], title: str = "Conversation History") -> Panel:
         """Renders a list of messages into a single panel (less used now with live updates)."""
         # This method might be less relevant if get_display_renderable directly builds the conversation.
@@ -737,18 +747,8 @@ class CLIRenderer:
         return table
         
     def render_error(self, error_message: str, details: Optional[str] = None) -> Panel:
-        """Renders an error message in a styled panel."""
-        error_text = Text(error_message, style=THEME_COLORS.get("error", "bold red"))
-        if details:
-            error_text.append("\n\nDetails:\n", style="bold")
-            error_text.append(details, style="") # Default style for details
-            
-        return Panel(
-            error_text, 
-            title="❌ Error", 
-            border_style=THEME_COLORS.get("error", "bold red"),
-            padding=(1,1) # Add some padding for errors
-        )
+        """Use unified renderer for error messages"""
+        return self.renderer.render_error(error_message, details=details)
 
     def reset_response_area(self) -> None:
         """Reset streaming state and other UI state."""
@@ -787,4 +787,4 @@ class CLIRenderer:
 # The Typer app and test execution logic is removed from this file.
 # This file now only contains the CLIRenderer class with enhanced rendering capabilities.
 # To test this, a separate script (like test_ui.py or the main cli.py)
-# would import CLIRenderer, instantiate it, and use its methods.
\ No newline at end of file
+# would import CLIRenderer, instantiate it, and use its methods.
