@@ -13,52 +13,36 @@ import tiktoken  # type: ignore
 # TODO: greatly simplify api_client while maintaining full functionality
 # TODO: add streaming support # Done for quite a while. 
 # TODO: add support for images, files, audio, and video
-from litellm import acompletion, completion, token_counter, cost_per_token, completion_cost # Keep litellm imports for now? maybe remove acompletion/completion direct use # type: ignore
+# Lazy import litellm to avoid 1+ second import time overhead
+# from litellm import acompletion, completion, token_counter, cost_per_token, completion_cost
 from PIL import Image  # type: ignore
 
 from .model_config import ModelConfig
 from .adapters import get_adapter # Keep for native preference
-from .litellm_gateway import LiteLLMGateway # Import the gateway
-from .openrouter_gateway import OpenRouterGateway # Import the OpenRouter gateway
+# Lazy import gateways to avoid import overhead
+# from .litellm_gateway import LiteLLMGateway
+# from .openrouter_gateway import OpenRouterGateway
 # from penguin.llm.provider_adapters import get_provider_adapter # Seems unused
 
 logger = logging.getLogger(__name__)
 
 
 def load_config() -> Dict[str, Any]:
+    """Load Penguin config using central resolver, with safe fallback.
+
+    Delegates to penguin.config.load_config, which already checks:
+    - PENGUIN_CONFIG_PATH
+    - ~/.config/penguin/config.yml (or platform equivalent)
+    - dev/repo defaults
+    Returns an empty dict on failure instead of raising during import.
     """
-    Load the configuration from the config.yml file.
-
-    This function reads the config.yml file located one directory up 
-    (expected to be penguin/penguin/) from the current file's directory 
-    (penguin/penguin/llm/).
-    It uses the yaml library to parse the YAML content into a Python dictionary.
-
-    Returns:
-        Dict[str, Any]: A dictionary containing the configuration data.
-                        If 'model_configs' key is not present, it returns an empty dictionary for that key.
-
-    Raises:
-        FileNotFoundError: If the config.yml file is not found.
-        yaml.YAMLError: If there's an error parsing the YAML content.
-    """
-    # Path(__file__) is .../llm/api_client.py
-    # .parent is .../llm/
-    # .parent is .../ (the penguin/penguin/ directory)
-    config_path = Path(__file__).parent.parent / "config.yml" # Updated path
-    if not config_path.exists():
-        logger.error(f"Config file not found at expected location: {config_path}")
-        raise FileNotFoundError(f"Config file not found at {config_path}")
-        
     try:
-        with open(config_path) as config_file:
-            return yaml.safe_load(config_file)
-    except yaml.YAMLError as e:
-        logger.error(f"Error parsing config file {config_path}: {e}")
-        raise
+        from penguin.config import load_config as core_load_config
+        data = core_load_config()
+        return data if isinstance(data, dict) else {}
     except Exception as e:
-        logger.error(f"Unexpected error loading config file {config_path}: {e}")
-        raise
+        logger.warning(f"api_client.load_config fallback: {e}")
+        return {}
 
 
 # Load the model configurations from the config file
@@ -97,6 +81,8 @@ class APIClient:
         # --- Instantiate the correct handler ---
         if model_config.client_preference == 'litellm':
             try:
+                # Lazy import LiteLLMGateway to avoid import overhead
+                from .litellm_gateway import LiteLLMGateway
                 # LiteLLM gateway handles API keys/base internally based on model_config
                 self.client_handler = LiteLLMGateway(model_config)
                 self.logger.info(f"Using LiteLLMGateway for {model_config.model}")
@@ -106,6 +92,8 @@ class APIClient:
 
         elif model_config.client_preference == 'openrouter':
             try:
+                # Lazy import OpenRouterGateway to avoid import overhead
+                from .openrouter_gateway import OpenRouterGateway
                 # Initialize OpenRouter gateway with model_config
                 self.client_handler = OpenRouterGateway(model_config)
                 self.logger.info(f"Using OpenRouterGateway for {model_config.model}")
