@@ -61,6 +61,15 @@ class OpenRouterGateway:
         self.site_url = site_url or os.getenv("OPENROUTER_SITE_URL")
         self.site_title = site_title or os.getenv("OPENROUTER_SITE_TITLE", "Penguin_AI_Agent")
 
+        # Simple telemetry counters
+        self._telemetry: Dict[str, Any] = {
+            "interrupts": 0,
+            "streamed_bytes": 0,
+        }
+        # Tool-call accumulation for SSE
+        self._tool_call_acc: Dict[str, Any] = {"name": None, "arguments": ""}
+        self._last_tool_call: Optional[Dict[str, Any]] = None
+
         # --- API Key Handling ---
         api_key = model_config.api_key or os.getenv("OPENROUTER_API_KEY")
         if not api_key:
@@ -156,6 +165,88 @@ class OpenRouterGateway:
                 # Not a list content, keep message as is
                 processed_messages.append(message)
         return processed_messages
+    
+    def _contains_penguin_action_tags(self, content: str) -> bool:
+        """
+        Check if content contains any Penguin action tags using the same logic as parser.py
+        """
+        try:
+            # Import here to avoid circular imports
+            from penguin.utils.parser import ActionType
+            import re
+            
+            # Generate pattern from ActionType enum (exactly like parser.py does)
+            action_tag_pattern = "|".join([action_type.value for action_type in ActionType])
+            # Use same regex pattern as parser: full tag pairs only, case-insensitive but strict
+            action_tag_regex = f"<({action_tag_pattern})>.*?</\\1>"
+            
+            return bool(re.search(action_tag_regex, content, re.DOTALL | re.IGNORECASE))
+        except ImportError:
+            # Fallback to basic check if import fails
+            return any(f"<{tag}>" in content.lower() and f"</{tag}>" in content.lower() 
+                      for tag in ['execute', 'search', 'memory_search'])
+    
+    def _clean_conversation_format(self, messages: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
+        """
+        Reformat conversation to be compatible with OpenAI SDK while preserving all content.
+        
+        This aggressively converts all tool calling to plain text format to avoid
+        OpenAI SDK 1.99+ tool call validation errors, while preserving all message content.
+        """
+        reformatted_messages = []
+        
+        for message in messages:
+            reformatted_message = message.copy()
+            
+            # Handle content field
+            if isinstance(message.get('content'), str):
+                content = message['content']
+                
+                # Clean up orphaned tool call references that could cause validation errors
+                if 'call_' in content and ('tool_calls' not in message):
+                    # Replace call_id references with plain text to avoid SDK validation
+                    import re
+                    content = re.sub(r'call_[a-zA-Z0-9_-]+', '[tool-call-reference]', content)
+                    self.logger.debug(f"Reformatted tool call references in message")
+                
+                # Check for XML action tags - they're Penguin's tool system and should be preserved
+                if self._contains_penguin_action_tags(content):
+                    self.logger.debug(f"Preserving Penguin XML action tags in message: {content[:100]}...")
+                
+                reformatted_message['content'] = content
+            
+            # AGGRESSIVE FIX: Convert ALL tool calling to plain text to avoid validation errors
+            # This prevents the "No tool call found" error by not sending tool call formats at all
+            
+            if message.get('role') == 'assistant' and 'tool_calls' in message:
+                # Convert assistant message with tool_calls to plain assistant message
+                self.logger.debug(f"Converting assistant message with tool_calls to plain text")
+                # Remove tool_calls field and keep content as-is (it already has the action tags)
+                reformatted_message = {
+                    'role': 'assistant',
+                    'content': message.get('content', '')
+                }
+                # Copy other fields but exclude tool_calls
+                for key, value in message.items():
+                    if key not in ['role', 'content', 'tool_calls']:
+                        reformatted_message[key] = value
+            
+            elif message.get('role') == 'tool':
+                # Convert ALL tool messages to assistant messages with prefixed content
+                self.logger.debug(f"Converting tool result message to assistant message")
+                reformatted_message = {
+                    'role': 'assistant', 
+                    'content': f"[Tool Result] {message.get('content', '')}"
+                }
+                # Copy other fields but exclude tool_call_id
+                for key, value in message.items():
+                    if key not in ['role', 'content', 'tool_call_id']:
+                        reformatted_message[key] = value
+            
+            reformatted_messages.append(reformatted_message)
+        
+        self.logger.debug(f"Reformatted conversation: {len(messages)} messages processed, all tool calling converted to plain text")
+        return reformatted_messages
 
     async def get_response(
         self,
@@ -205,15 +296,17 @@ class OpenRouterGateway:
             self.logger.warning("Streaming requested/configured but no stream_callback provided. Falling back to non-streaming mode.")
             use_streaming = False
 
-        # --- Process messages for vision --- 
+        # --- Process messages for vision and reformat conversation --- 
         try:
             processed_messages = await self._process_messages_for_vision(messages)
+            # Reformat conversation to be compatible with OpenAI SDK while preserving content
+            processed_messages = self._clean_conversation_format(processed_messages)
         except Exception as e:
             # error_context = {'request_id': request_id, 'phase': 'vision_processing', 'messages_count': len(messages)}
             # debug_error(e, error_context)
-            self.logger.error(f"Error processing messages for vision: {e}", exc_info=True)
+            self.logger.error(f"Error processing messages for vision and conversation format: {e}", exc_info=True)
             return f"[Error: Failed to process message content - {str(e)}]"
-        # --- End vision processing ---
+        # --- End vision and conversation processing ---
 
         # --- Reasoning tokens configuration ---
         reasoning_config = self.model_config.get_reasoning_config()
@@ -228,21 +321,21 @@ class OpenRouterGateway:
             **kwargs # Pass through other arguments like tools
         }
         
-        # Add include_reasoning parameter if reasoning is enabled
+        # Add new unified reasoning parameter if reasoning is enabled
         if reasoning_config:
-            request_params["include_reasoning"] = True
-            self.logger.info(f"[OpenRouterGateway] Enabling reasoning tokens with include_reasoning=True")
-        
-        # Handle reasoning configuration carefully to avoid SDK compatibility issues
-        use_direct_api = False
-        if reasoning_config:
-            try:
-                # Try to add reasoning config - if SDK doesn't support it, we'll catch the error
+            # Use new unified reasoning format instead of include_reasoning
+            if isinstance(reasoning_config, dict):
                 request_params["reasoning"] = reasoning_config
-                self.logger.info(f"[OpenRouterGateway] Adding reasoning config: {reasoning_config}")
-            except Exception as e:
-                self.logger.warning(f"[OpenRouterGateway] Reasoning config not supported by SDK, will use direct API: {e}")
-                use_direct_api = True
+                self.logger.info(f"[OpenRouterGateway] Using new reasoning config: {reasoning_config}")
+            else:
+                # Fallback to simple enabled format for backwards compatibility
+                request_params["reasoning"] = {"enabled": True}
+                self.logger.info(f"[OpenRouterGateway] Using basic reasoning config with enabled=True")
+        
+        # Handle reasoning configuration - always use direct API for reasoning
+        use_direct_api = bool(reasoning_config)
+        if reasoning_config:
+            self.logger.info(f"[OpenRouterGateway] Reasoning enabled, will use direct API call to bypass SDK limitations")
             
         # Filter out None values for cleaner API calls
         request_params = {k: v for k, v in request_params.items() if v is not None}
@@ -258,18 +351,17 @@ class OpenRouterGateway:
 
         full_response_content = ""
         full_reasoning_content = ""
+        
+        # Use direct API call if reasoning is enabled to avoid SDK compatibility issues
+        if use_direct_api:
+            self.logger.debug(f"[OpenRouterGateway] Using direct API call for reasoning support")
+            return await self._direct_api_call_with_reasoning(
+                request_params, reasoning_config, use_streaming, stream_callback
+            )
+        
+        # Use OpenAI SDK for non-reasoning requests
         try:
-            # Try the API call with reasoning parameter
             completion = await self.client.chat.completions.create(**request_params)
-        except TypeError as e:
-            if "reasoning" in str(e) and reasoning_config:
-                # SDK doesn't support reasoning parameter, try without it and use httpx fallback
-                self.logger.warning(f"[OpenRouterGateway] SDK doesn't support reasoning parameter, falling back to direct API call")
-                return await self._direct_api_call_with_reasoning(
-                    request_params, reasoning_config, use_streaming, stream_callback
-                )
-            else:
-                raise
         except Exception:
             raise
             
@@ -341,6 +433,10 @@ class OpenRouterGateway:
                         
                         if new_content_segment:
                             _gateway_accumulated_content += new_content_segment
+                            try:
+                                self._telemetry["streamed_bytes"] += len(new_content_segment.encode("utf-8"))
+                            except Exception:
+                                pass
                             # debug_stream_chunk(request_id, {'chunk': new_content_segment, 'type': 'content'}, "content")
                             if stream_callback:
                                 try:
@@ -352,9 +448,56 @@ class OpenRouterGateway:
                         
                         full_response_content = _gateway_accumulated_content
 
+                        # Interrupt streaming when a complete Penguin action tag is detected
+                        try:
+                            if getattr(self.model_config, "interrupt_on_action", False):
+                                if self._contains_penguin_action_tags(full_response_content):
+                                    self.logger.info("[OpenRouterGateway] Interrupting stream on detected Penguin action tag (SDK path)")
+                                    try:
+                                        self._telemetry["interrupts"] += 1
+                                    except Exception:
+                                        pass
+                                    return full_response_content
+                        except Exception as _int_err:
+                            self.logger.debug(f"[OpenRouterGateway] interrupt_on_action check failed: {_int_err}")
+
                     elif tool_calls_delta:
                          self.logger.debug(f"[OpenRouterGateway] Received tool_calls delta: {tool_calls_delta}.")
-                         # Tool call streaming logic would go here if needed for external callback
+                         # Accumulate name/arguments from delta (best-effort)
+                         try:
+                             tc0 = None
+                             if isinstance(tool_calls_delta, (list, tuple)) and tool_calls_delta:
+                                 tc0 = tool_calls_delta[0]
+                             if tc0 is not None:
+                                 fn = getattr(tc0, "function", None) if not isinstance(tc0, dict) else tc0.get("function")
+                                 if fn is not None:
+                                     name = getattr(fn, "name", None) if not isinstance(fn, dict) else fn.get("name")
+                                     args_delta = getattr(fn, "arguments", None) if not isinstance(fn, dict) else fn.get("arguments")
+                                     if name and not self._tool_call_acc.get("name"):
+                                         self._tool_call_acc["name"] = name
+                                     if isinstance(args_delta, str) and args_delta:
+                                         self._tool_call_acc["arguments"] += args_delta
+                         except Exception as _acc_err:
+                             self.logger.debug(f"[OpenRouterGateway] tool_call accumulation failed: {_acc_err}")
+                         # Interrupt on tool_call if enabled
+                         try:
+                             if getattr(self.model_config, "interrupt_on_tool_call", False):
+                                 self.logger.info("[OpenRouterGateway] Interrupting stream on tool_call delta (SDK path)")
+                                 # Snapshot last tool call
+                                 try:
+                                     self._last_tool_call = {
+                                         "name": self._tool_call_acc.get("name"),
+                                         "arguments": self._tool_call_acc.get("arguments", ""),
+                                     }
+                                 except Exception:
+                                     self._last_tool_call = None
+                                 try:
+                                     self._telemetry["interrupts"] += 1
+                                 except Exception:
+                                     pass
+                                 return _gateway_accumulated_content
+                         except Exception as _tool_int_err:
+                             self.logger.debug(f"[OpenRouterGateway] interrupt_on_tool_call check failed: {_tool_int_err}")
                     else:
                         self.logger.debug(f"[OpenRouterGateway] Chunk {chunk_index-1} had no text/reasoning/tool delta.")
 
@@ -460,9 +603,9 @@ class OpenRouterGateway:
         direct_params = request_params.copy()
         extra_headers = direct_params.pop("extra_headers", {})
         
-        # Add reasoning configuration
+        # Add reasoning configuration using new unified format
         direct_params["reasoning"] = reasoning_config
-        direct_params["include_reasoning"] = True
+        # Remove include_reasoning - now using unified reasoning parameter
         
         # Prepare headers
         headers = {
@@ -505,7 +648,7 @@ class OpenRouterGateway:
         
         async with client.stream("POST", url, headers=headers, json=params) as response:
             if response.status_code != 200:
-                error_text = await response.atext()
+                error_text = (await response.aread()).decode()
                 self.logger.error(f"Direct API call failed with status {response.status_code}: {error_text}")
                 return f"[Error: API call failed with status {response.status_code}]"
             
@@ -542,11 +685,62 @@ class OpenRouterGateway:
                                 self.logger.debug("Reasoning phase complete, switching to content")
                             
                             full_content += content_delta
+                            try:
+                                self._telemetry["streamed_bytes"] += len(content_delta.encode("utf-8"))
+                            except Exception:
+                                pass
                             if stream_callback:
                                 try:
                                     await stream_callback(content_delta, "assistant")
                                 except Exception as cb_err:
                                     self.logger.error(f"Error in content callback: {cb_err}")
+                            # Interrupt streaming when a complete Penguin action tag is detected
+                            try:
+                                if getattr(self.model_config, "interrupt_on_action", False):
+                                    if self._contains_penguin_action_tags(full_content):
+                                        self.logger.info("[OpenRouterGateway] Interrupting stream on detected Penguin action tag (Direct API path)")
+                                        try:
+                                            self._telemetry["interrupts"] += 1
+                                        except Exception:
+                                            pass
+                                        break
+                            except Exception as _int_err:
+                                self.logger.debug(f"[OpenRouterGateway] interrupt_on_action check failed: {_int_err}")
+                        # Handle tool_calls in direct SSE (Responses/OpenAI compatible)
+                        try:
+                            tool_calls_delta = getattr(delta, "tool_calls", None) if hasattr(delta, "tool_calls") else delta.get("tool_calls")
+                            if tool_calls_delta and getattr(self.model_config, "interrupt_on_tool_call", False):
+                                # Accumulate information
+                                try:
+                                    tc0 = None
+                                    if isinstance(tool_calls_delta, (list, tuple)) and tool_calls_delta:
+                                        tc0 = tool_calls_delta[0]
+                                    if tc0 is not None:
+                                        fn = getattr(tc0, "function", None) if not isinstance(tc0, dict) else tc0.get("function")
+                                        if fn is not None:
+                                            name = getattr(fn, "name", None) if not isinstance(fn, dict) else fn.get("name")
+                                            args_delta = getattr(fn, "arguments", None) if not isinstance(fn, dict) else fn.get("arguments")
+                                            if name and not self._tool_call_acc.get("name"):
+                                                self._tool_call_acc["name"] = name
+                                            if isinstance(args_delta, str) and args_delta:
+                                                self._tool_call_acc["arguments"] += args_delta
+                                except Exception as _acc_err2:
+                                    self.logger.debug(f"[OpenRouterGateway] tool_call accumulation failed: {_acc_err2}")
+                                self.logger.info("[OpenRouterGateway] Interrupting stream on tool_call delta (Direct API path)")
+                                try:
+                                    self._last_tool_call = {
+                                        "name": self._tool_call_acc.get("name"),
+                                        "arguments": self._tool_call_acc.get("arguments", ""),
+                                    }
+                                except Exception:
+                                    self._last_tool_call = None
+                                try:
+                                    self._telemetry["interrupts"] += 1
+                                except Exception:
+                                    pass
+                                break
+                        except Exception as _tool_int_err2:
+                            self.logger.debug(f"[OpenRouterGateway] interrupt_on_tool_call check failed: {_tool_int_err2}")
                         
                     except json.JSONDecodeError as e:
                         self.logger.warning(f"Failed to parse SSE data: {data_str[:100]}... Error: {e}")
@@ -555,6 +749,23 @@ class OpenRouterGateway:
         self.logger.info(f"Direct streaming call completed. Reasoning: {len(full_reasoning)} chars, Content: {len(full_content)} chars")
         return full_content
 
+    def get_telemetry(self) -> Dict[str, Any]:
+        """Return simple telemetry counters for diagnostics."""
+        try:
+            return dict(self._telemetry)
+        except Exception:
+            return {"interrupts": 0, "streamed_bytes": 0}
+
+    def get_and_clear_last_tool_call(self) -> Optional[Dict[str, Any]]:
+        """Return last detected tool_call (name, arguments) and clear accumulators."""
+        try:
+            data = self._last_tool_call
+            self._last_tool_call = None
+            self._tool_call_acc = {"name": None, "arguments": ""}
+            return data
+        except Exception:
+            return None
+
     async def _handle_non_streaming_response(
         self,
         client: httpx.AsyncClient,
@@ -578,8 +789,9 @@ class OpenRouterGateway:
             choice = data.get("choices", [{}])[0]
             message = choice.get("message", {})
             
-            content = message.get("content", "")
-            reasoning = message.get("reasoning", "")
+            # Some providers include keys with explicit None values; coalesce to empty strings
+            content = message.get("content") or ""
+            reasoning = message.get("reasoning") or ""
             
             # If we have reasoning and a callback, emit it
             if reasoning and stream_callback:
