@@ -3,19 +3,48 @@ Defines the syntax for actions Penguin can take and provides critical usage guid
 especially regarding safety and verification.
 """
 
-ACTION_SYNTAX = """
-## Action Syntax
+# Import shared formatting rules
+from penguin.prompt_workflow import SAFETY_RULES, CODE_FORMATTING_RULES, TOOL_RESULT_HANDLING
 
-**Core Principles Reminder:**
-- **Verify BEFORE Acting:** Check state before modifying.
-- **Safety First:** Avoid overwrites, confirm destructive actions.
-- **Acknowledge Results:** Confirm outcomes from the previous message before proceeding.
+_ACTION_SYNTAX_TEMPLATE = r"""
+## Action Syntax
 
 ---
 
-### Code Execution (`<execute>`)
+### Output Formatting (MANDATORY)
 
-Executes Python code within an IPython environment in the workspace. Use this for file operations, checks, data processing, etc.
+To ensure perfect TUI rendering, follow all of the below:
+
+1) Fence all multi-line content with triple backticks and a language:
+   - `python`, `javascript`, `json`, `bash` (or `sh`), `toml`, `yaml`, `ini`, `diff`, `xml`.
+   - For ActionTag/XML-like content (`<execute>`, `<apply_diff>`, `<enhanced_*`, `<tool>`, `<action>`), use the custom language `actionxml`.
+
+2) Never inline multi-line code without fences. Do not mix narrative text inside fenced JSON/code.
+
+3) Tool/Action display pattern:
+   - Start with a short caption line (what this block is), then a single fenced block with the content.
+   - Avoid duplication: if the tool will output the full content, summarize briefly and rely on the tool block; do not re-echo the same code elsewhere.
+
+4) Diff output:
+   - Use `diff` fenced blocks with standard unified headers and hunks:```diff
+--- a/path
++++ b/path
+@@ -1,3 +1,4 @@
+ old
++new
+```
+
+5) Headline then details:
+   - Begin with a one-line summary of the action/result; put the body in a fenced block immediately after.
+
+6) Long outputs:
+   - Prefer concise assistant messages. Large text should come from tool outputs; use a brief preview and say “See tool result” if necessary.
+
+---
+
+### Code Execution (`<execute>` and `<execute_command>`)
+
+Execute code/commands in the workspace. The execution environment is typically detected from project structure.
 
 **Action Result Handling:**
 Results (stdout, stderr, errors) appear in the *next* system message. **You MUST wait for this message.** Your *next* response must:
@@ -23,22 +52,20 @@ Results (stdout, stderr, errors) appear in the *next* system message. **You MUST
 2.  **Verify** the outcome (e.g., check file existence/content).
 3.  **Proceed** based *only* on the verified outcome.
 
-**Example (Safe File Creation):**
+**Example 1: Python (when project has *.py files, pyproject.toml, etc.):**
 ```python
 # <execute>
 import os
 from pathlib import Path
 
-file_path = Path('discord_clone/backend/src/controllers/auth.controller.js')
+file_path = Path('src/main.py')
 print(f"Checking existence of: {file_path}")
 
 if not file_path.exists():
     print(f"File does not exist. Creating...")
-    # Ensure parent directory exists
     file_path.parent.mkdir(parents=True, exist_ok=True)
     print(f"Ensured directory {file_path.parent} exists.")
-    # Write content (replace with actual content variable)
-    content_to_write = "console.log('Auth Controller');"
+    content_to_write = "def main():\n    print('Hello')\n"
     try:
         file_path.write_text(content_to_write, encoding='utf-8')
         print(f"Successfully created and wrote to {file_path}")
@@ -49,19 +76,37 @@ else:
 # </execute>
 ```
 
-**CRITICAL SAFETY WARNINGS for `<execute>`:**
--   **NEVER use `open(path, 'w')` or `Path(path).write_text()` without FIRST checking `os.path.exists(path)` or `Path(path).exists()`.**
--   If the file exists, **DO NOT OVERWRITE** unless explicitly instructed or you have stated a clear, verified reason (e.g., content is incorrect based on a prior read). Prefer reading, modifying specific lines, or creating backups if modification is needed. **State your intent clearly before modifying existing files.**
--   Be equally cautious with `os.remove`, `shutil.rmtree`, `os.makedirs(exist_ok=False)`, `Path.unlink()`, `shutil.rmtree()`. Verify targets and confirm intent.
--   Use `pathlib` for safer and easier path manipulation and checks (`Path.exists()`, `Path.is_file()`, `Path.read_text()`, `Path.write_text()`, `Path.mkdir()`).
--   Ensure parent directories exist before writing files (`file_path.parent.mkdir(parents=True, exist_ok=True)`).
+**Example 2: JavaScript/Node (when project has package.json):**
+```javascript
+// Use execute_command for Node.js
+<execute_command>
+node -e "const fs = require('fs'); const path = 'src/index.js'; if (fs.existsSync(path)) { console.log('File exists:', path); } else { console.log('File does not exist:', path); }"
+</execute_command>
+```
 
-**Other Notes for `<execute>`:**
--   Keep scripts **short and focused** (one logical operation per block).
--   Write long files **incrementally**, verifying each chunk.
--   Include `print()` statements for status updates and verification points.
--   Use `try...except` for error handling within scripts.
--   Manage paths explicitly using `os.getcwd()`, `Path.cwd()`, and `os.path.join()` or `pathlib` operators. `cd` does not persist.
+**Example 3: Shell Commands (language-agnostic checks):**
+```actionxml
+<execute_command>ls -la src/ && find . -name "*.config.*" -type f</execute_command>
+```
+
+**Example 4: Rust (when project has Cargo.toml):**
+```actionxml
+<execute_command>cargo check --verbose</execute_command>
+```
+
+**CRITICAL SAFETY WARNINGS (all languages):**
+-   **NEVER write to files without FIRST checking existence**
+-   For Python: use `Path(path).exists()` before `Path(path).write_text()`
+-   For Shell: use `[ -f path ]` or `ls path` before redirecting output
+-   If file exists, **DO NOT OVERWRITE** without verification. Prefer `apply_diff` or backups.
+-   Be cautious with destructive operations (`rm`, `shutil.rmtree`, file deletion)
+-   Verify targets and confirm intent before modifying/deleting
+
+**Other Notes:**
+-   Keep scripts **short and focused** (one logical operation per block)
+-   Include status output for verification (print, echo, logging)
+-   Use `try...except` (Python) or error handling for robust scripts
+-   For multi-language projects: detect language first via project files
 
 ---
 
@@ -122,21 +167,71 @@ Executes a shell command in the workspace root. **Use sparingly and cautiously.*
 
 1.  **Apply Unified Diff (`<apply_diff>`):**
     `<apply_diff>file_path:diff_content:backup</apply_diff>`
-    -   **Applies unified diff format** to make precise edits.
-    -   **Targets specific lines** - does not just append content.
-    -   Creates backup by default.
-    -   Example with diff content:
-    ```
+    -   **Edit a single file** with unified diff format for precise line-based changes.
+    -   **Applies immediately** (no dry-run mode). Backups are created by default (`backup=true`).
+    -   Tip: If your diff content contains colons, omit the optional `:backup` parameter and rely on the default.
+
+    **Example (applies immediately):**
+    ```actionxml
     <apply_diff>src/main.py:--- a/src/main.py
     +++ b/src/main.py
     @@ -1,3 +1,4 @@
      def hello():
-    +    ""Say hello.""
+    +    \"\"\"Say hello.\"\"\"
          print("Hello")
-    :true</apply_diff>
+    </apply_diff>
     ```
 
-2.  **Pattern-Based Editing (`<edit_with_pattern>`):**
+2.  **Multi-File Editing (`<multiedit>`):**
+    `<multiedit>content</multiedit>`
+    -   **Apply multiple diffs atomically** - all succeed or none are applied.
+    -   **DRY-RUN BY DEFAULT** - shows what would change without applying.
+    -   **Creates automatic backups** for all modified files.
+    -   Supports **per-file block format only** (standard unified multi-file patches are not accepted by this tool).
+    -   Add `apply=true` as the first line inside the tag to actually apply.
+    
+    **Per-File Blocks (supported):**
+    ```actionxml
+    <multiedit>
+    path/to/file1.py:
+    --- a/path/to/file1.py
+    +++ b/path/to/file1.py
+    @@ -1,2 +1,3 @@
+     import os
+    +from pathlib import Path
+     print("hi")
+    
+    path/to/new_file.txt:
+    @@ -0,0 +1,2 @@
+    +hello
+    +world
+    </multiedit>
+    ```
+    
+    **Apply Mode Example:**
+    ```actionxml
+    <multiedit>
+    apply=true
+    src/config.py:
+    @@ -10,11 +10,12 @@
+     DEBUG = False
+    +LOG_LEVEL = "INFO"
+     PORT = 8080
+    
+    src/main.py:
+    @@ -1,2 +1,3 @@
+    +#!/usr/bin/env python3
+     import config
+    </multiedit>
+    ```
+    
+    **Notes:**
+    -   All edits are validated before any are applied
+    -   Failed validation shows errors for all problematic edits
+    -   Automatic rollback if any edit fails during application
+    -   Preserves file permissions and attributes
+
+3.  **Pattern-Based Editing (`<edit_with_pattern>`):**
     `<edit_with_pattern>file_path:search_pattern:replacement:backup</edit_with_pattern>`
     -   **Uses regex patterns** for find-and-replace operations.
     -   Safer than manual string replacement.
@@ -165,15 +260,18 @@ Executes a shell command in the workspace root. **Use sparingly and cautiously.*
 
 Use for information retrieval. Acknowledge results before acting.
 
-1.  **Web Search (`<perplexity_search>`):**
+1.  **Grep-like Search (`<search>`):**
+    `<search>pattern</search>`
+    -   Project-local, fast regex search over logs and files. Use `|` to OR patterns.
+    -   Example: `<search>def\s+my_function|class\s+MyClass</search>`
+
+2.  **Web Search (`<perplexity_search>`):**
     `<perplexity_search>query:max_results</perplexity_search >`
     -   External, current info (docs, concepts, news). Max 5 results.
     -   Example: `<perplexity_search>python requests library usage:3</perplexity_search >`
 
-2.  **Codebase Search (`<workspace_search>`):**
-    `<workspace_search>query:max_results</workspace_search >`
-    -   **Use FIRST** for finding code (functions, classes, variables) in the project.
-    -   Example: `<workspace_search>class UserSchema:5</workspace_search >`
+3.  **Codebase Search (`<workspace_search>`):**
+    -   Currently disabled for performance reasons in this build. Use `<search>` and `<find_files_enhanced>` instead.
 
 3.  **Memory Search (`<memory_search>`):**
     `<memory_search>query:k:memory_type:categories</memory_search >`
@@ -211,6 +309,17 @@ Manage long-running background processes.
 
 ---
 
+### File Editing Best Practices
+
+**Key Principles:**
+-   **Always dry-run first** - Review changes before applying
+-   **Verify state before editing** - Read files to understand current content
+-   **Use appropriate tool** - `apply_diff` for single files, `multiedit` for multiple
+-   **Keep scripts short** - Acknowledge results, continue through recoverable errors
+-   **Enhanced tools available** - File ops, search, and browser tools handle backups/safety automatically
+
+---
+
 ### Memory Management (`add_*_note`)
 
 Preserve context using notes.
@@ -234,6 +343,72 @@ Track high-level plan progress.
 
 ---
 
+### Git & GitHub Operations
+
+For Git operations and PR creation, use `<execute>` with Python code:
+
+**Available libraries and credentials:**
+- `git` command-line tool (subprocess)
+- `PyGithub` library for GitHub API operations
+- GitHub App authentication is configured when these env vars are set:
+  - `GITHUB_APP_ID`, `GITHUB_APP_INSTALLATION_ID`, `GITHUB_APP_PRIVATE_KEY_PATH`
+  - `GITHUB_REPOSITORY` (target repo, e.g., "owner/repo")
+
+**Example PR creation workflow:**
+```python
+# <execute>
+import subprocess
+import jwt
+import time
+from github import Auth, Github, GithubIntegration
+import os
+
+# 1. Clone repo
+subprocess.run(['git', 'clone', 'https://github.com/owner/repo.git', 'repo-dir'])
+
+# 2. Create branch, make changes, commit
+os.chdir('repo-dir')
+subprocess.run(['git', 'checkout', '-b', 'feature-branch'])
+# ... make file changes ...
+subprocess.run(['git', 'add', '.'])
+subprocess.run(['git', 'commit', '-m', 'feat: add feature'])
+
+# 3. Get GitHub App installation token
+app_id = os.getenv('GITHUB_APP_ID')
+installation_id = os.getenv('GITHUB_APP_INSTALLATION_ID')
+key_path = os.getenv('GITHUB_APP_PRIVATE_KEY_PATH')
+
+with open(key_path) as f:
+    private_key = f.read()
+    
+auth = Auth.AppAuth(app_id, private_key)
+installation_auth = auth.get_installation_auth(int(installation_id))
+token = installation_auth.token
+
+# 4. Push with token
+remote_url = f"https://x-access-token:{token}@github.com/owner/repo.git"
+subprocess.run(['git', 'remote', 'set-url', 'origin', remote_url])
+subprocess.run(['git', 'push', 'origin', 'feature-branch'])
+
+# 5. Create PR via API
+github = Github(auth=installation_auth)
+repo = github.get_repo('owner/repo')
+pr = repo.create_pull(
+    title="PR Title",
+    body="PR Description",
+    head="feature-branch",
+    base="main"
+)
+print(f"PR created: {pr.html_url}")
+# </execute>
+```
+
+**Notes:**
+- This approach gives you full control and visibility into each step
+- Works with any Git hosting (GitHub, GitLab, etc.) by swapping the API client
+- All GitHub App credentials are automatically available via env vars
+
+---
 
 **Notes:**
 -   Navigate first. Use specific selectors.
@@ -249,6 +424,12 @@ Enhanced browser control without WebDriver dependencies, better for sites with a
 
 -   `<pydoll_browser_navigate>URL</pydoll_browser_navigate >`
 -   `<pydoll_browser_interact>action:selector[:selector_type][:text]</pydoll_browser_interact >` (actions: `click`, `input`, `submit`, selector_types: `css`, `xpath`, `id`, `class_name`)
+-   `<pydoll_browser_scroll>mode:args</pydoll_browser_scroll >` (modes: `to`, `by`, `page`, `element`)
+    - Examples:
+      - `<pydoll_browser_scroll>to:bottom</pydoll_browser_scroll>`
+      - `<pydoll_browser_scroll>page:down:2</pydoll_browser_scroll>`
+      - `<pydoll_browser_scroll>by:1200:0:1</pydoll_browser_scroll>`
+      - `<pydoll_browser_scroll>element:#results:css:smooth</pydoll_browser_scroll>`
 -   `<pydoll_browser_screenshot></pydoll_browser_screenshot >`
 -   `<pydoll_debug_toggle>[on|off]</pydoll_debug_toggle >` (Enable/disable detailed PyDoll logging and outputs)
 
@@ -272,9 +453,124 @@ Enhanced browser control without WebDriver dependencies, better for sites with a
 -   Helps diagnose issues with selectors or page navigation
 -   Toggle with: `<pydoll_debug_toggle>on</pydoll_debug_toggle >` or `<pydoll_debug_toggle>off</pydoll_debug_toggle >`
 
+### Inter-Agent Messaging (`<send_message>`)
+
+Use this action when agents need to coordinate directly or broadcast updates to a shared room. The body must be valid JSON. Required fields:
+
+- `content`: The plain-text message
+- `target` **or** `targets`: A string recipient (`"planner"`, `"qa"`, etc.) or a list of recipients. Omit to reach the human operator.
+
+Optional fields:
+
+- `channel`: Logical room identifier (e.g. `"dev-room"`). Use this so other agents can filter transcripts.
+- `message_type`: Defaults to `"message"`; useful for `"status"`, `"action"`, etc.
+- `metadata`: Arbitrary key/value payload carried alongside the message.
+- `sender`: Custom sender label (defaults to the current agent).
+
+**Examples:**
+
+```actionxml
+<send_message>{
+  "target": "implementer",
+  "content": "Planner has prepared the remediation plan. See dev-room for details.",
+  "channel": "dev-room",
+  "metadata": {"plan_version": 2}
+}</send_message>
+```
+
+Broadcasting to multiple recipients:
+
+```actionxml
+<send_message>{
+  "targets": ["planner", "qa"],
+  "content": "Implementation complete. Please review.",
+  "channel": "dev-room",
+  "message_type": "status"
+}</send_message>
+```
+
+Ping the human operator when assistance is required:
+
+```actionxml
+<send_message>{
+  "content": "Need clarification on acceptance criteria for the telemetry dashboard.",
+  "channel": "dev-room"
+}</send_message>
+```
+
+The message is routed through the MessageBus, recorded in each agent's transcript with channel metadata, and surfaced in telemetry summaries.
+
+---
+
+### Sub-Agent Tools (Agents-as-Tools)
+
+Use these to manage child agents directly. Defaults are conservative: sub-agents are created with isolated session and isolated context-window. Initial SYSTEM/CONTEXT is copied once; subsequent turns diverge. No permission engine is enforced; any `default_tools` provided are recorded as metadata only.
+
+1) Spawn Sub-Agent (`<spawn_sub_agent>`)
+
+Body must be JSON. Fields:
+- `id` (required): child agent id
+- `parent` (optional): parent agent id (defaults to current)
+- `persona`, `system_prompt` (optional)
+- `share_session` (default false) – share parent session
+- `share_context_window` (default false) – share parent context window
+- `shared_cw_max_tokens` (optional int) – clamp when using isolated CW
+- `model_config_id`, `model_overrides`, `model_max_tokens` (optional)
+- `default_tools` (optional list) – recorded only; not enforced
+- `initial_prompt` (optional) – message sent to the child after spawn
+
+Example:
+```actionxml
+<spawn_sub_agent>{
+  "id": "researcher",
+  "parent": "primary",
+  "persona": "research",
+  "share_session": false,
+  "share_context_window": false,
+  "shared_cw_max_tokens": 512,
+  "initial_prompt": "Summarize docs in /docs"
+}</spawn_sub_agent>
+```
+
+2) Pause / Resume Sub-Agent
+
+```actionxml
+<stop_sub_agent>{"id": "researcher"}</stop_sub_agent>
+<resume_sub_agent>{"id": "researcher"}</resume_sub_agent>
+```
+
+3) Delegate Work to Existing Agent (`<delegate>`)
+
+Send a task to a specific agent and record a delegation event.
+
+```actionxml
+<delegate>{
+  "parent": "primary",
+  "child": "researcher",
+  "content": "Audit README for gaps",
+  "channel": "dev-room",
+  "metadata": {"priority": "high"}
+}</delegate>
+```
+
+Notes:
+- For general multi-agent chatter, use `<send_message>` with `target`/`targets` and optional `channel`.
+- Cross-parent channel rooms are a future consideration; prefer direct `target` for sub-agent flows today.
+
 ---
 """
 
+# Construct ACTION_SYNTAX with shared constants at the beginning
+ACTION_SYNTAX = f"""
+## Action Syntax
+
+{SAFETY_RULES}
+
+{TOOL_RESULT_HANDLING}
+
+{CODE_FORMATTING_RULES}
+
+""" + _ACTION_SYNTAX_TEMPLATE
 
 
 PLACEHOLDER = """
@@ -295,6 +591,8 @@ Description: Run code in the terminal, using iPython or shell/bash (depending on
 <process_list></process_list> - List processes
 <process_enter>name</process_enter> - Enter a process
 <process_send>name:message</process_send> - Send a message to a process
+<send_message>{"target": "planner", "content": "Update", "channel": "dev-room"}</send_message> - Send inter-agent messages (JSON body)
+<send_message>{"target": "planner", "content": "Update", "channel": "dev-room"}</send_message> - Send inter-agent messages (JSON body)
 
 # Enhanced File Operations (USE THESE INSTEAD OF RAW PYTHON FILE OPS)
 
@@ -304,7 +602,10 @@ Description: Run code in the terminal, using iPython or shell/bash (depending on
 <find_files_enhanced>pattern:search_path:include_hidden:file_type</find_files_enhanced> - Find files with glob patterns
 <enhanced_diff>file1:file2:semantic</enhanced_diff> - Compare files with semantic analysis
 <analyze_project>directory:include_external</analyze_project> - Analyze project structure with AST
-<apply_diff>file_path:diff_content:backup</apply_diff> - Apply unified diff to edit files precisely
+
+# File Editing (Dry-run by default, add apply=true to apply)
+<apply_diff>file_path:diff_content:backup</apply_diff> - Edit single file with unified diff (creates backups)
+<multiedit>content</multiedit> - Apply multiple diffs atomically (all succeed or none)
 <edit_with_pattern>file_path:search_pattern:replacement:backup</edit_with_pattern> - Edit files with regex patterns
 
 # Memory
@@ -413,4 +714,4 @@ Description: Run code in the terminal, using iPython or shell/bash (depending on
 #             tool_list.append(f"- {tool['name']} ({tool.get('author', 'Unknown')})")
 #             tool_list.append(f"  {tool['description']}")
     
-#     return "\n".join(tool_list)
\ No newline at end of file
+#     return "\n".join(tool_list)
