@@ -488,6 +488,37 @@ class RunMode:
                     task_data["context"]
                 )
                 
+                # Check if we got an LLM empty response error - should stop continuous mode
+                # TODO: Consider retrying the task/request, if need be, with a different model or prompt
+                if task_result.get("completion_type") == "llm_empty_response_error":
+                    logger.warning("LLM empty response error in continuous mode, stopping")
+                    await self._emit_event({
+                        "type": "message",
+                        "role": "system",
+                        "content": task_result.get("message", "LLM returned empty response"),
+                        "category": MessageCategory.ERROR
+                    })
+                    await self._emit_event({
+                        "type": "status",
+                        "status_type": "continuous_mode_ending",
+                        "data": {"reason": "llm_empty_response"}
+                    })
+                    self._shutdown_requested = True
+                    break
+                
+                # Check for other errors
+                if task_result.get("status") == "error" and task_result.get("completion_type") == "error":
+                    logger.error(f"Error during task execution: {task_result.get('message')}")
+                    # For generic errors, emit event but continue (may be transient)
+                    await self._emit_event({
+                        "type": "message",
+                        "role": "system",
+                        "content": f"Task error: {task_result.get('message')}. Continuing with next task...",
+                        "category": MessageCategory.ERROR
+                    })
+                    # Small delay before continuing to avoid rapid error loops
+                    await asyncio.sleep(2)
+                
                 # Check if we should exit continuous mode based on the result
                 if task_result.get("completion_type") == "user_specified":
                     msg_content = "RunMode: User-specified task completed, waiting for further instructions."
@@ -773,13 +804,26 @@ class RunMode:
 
                     if new_part:
                         # Forward the *new* part to Core so that the UI receives a
-                        # single, de-duplicated stream of text chunks.
+                        # single, de-duplicated stream of text chunks. Use the
+                        # standard "assistant" message_type so downstream renderers
+                        # (TUI/web) treat this as primary assistant content rather
+                        # than an opaque "text" subtype.
                         try:
-                            await self.core._handle_stream_chunk(new_part, message_type="text", role="assistant")
+                            await self.core._handle_stream_chunk(new_part, message_type="assistant", role="assistant")
                         except Exception:
                             pass  # Never break RunMode on UI errors
                     return  # Assistant handled – skip further processing
 
+                # ------------------------------------------------------------------
+                # 1b. Reasoning stream – route via Core so TUI renders sidebar
+                # ------------------------------------------------------------------
+                if message_type == "reasoning":
+                    try:
+                        await self.core._handle_stream_chunk(message, message_type="reasoning", role="assistant")
+                    except Exception:
+                        pass
+                    return
+
                 # ------------------------------------------------------------------
                 # 2. Tool output, errors, system notes – emit as regular events
                 # ------------------------------------------------------------------
@@ -829,6 +873,8 @@ class RunMode:
             # Call Engine.run_task – streaming enabled via Engine.settings;
             # assistant chunks arrive through *message_callback* above.
             # ------------------------------------------------------------------
+            agent_override = context.get("agent_id") if context else None
+            agent_role = context.get("agent_role") if context else None
             result = await self.core.engine.run_task(
                 task_prompt=task_prompt,
                 max_iterations=self.max_iterations,
@@ -837,7 +883,9 @@ class RunMode:
                 task_name=name,
                 completion_phrases=completion_phrases,
                 enable_events=True,
-                message_callback=message_callback
+                message_callback=message_callback,
+                agent_id=agent_override,
+                agent_role=agent_role,
             )
             
             # Ensure any active streaming message is finalised so UI gets a single completed panel
@@ -865,7 +913,37 @@ class RunMode:
             }
                 
         except Exception as e:
-            # Handle execution errors
+            # Handle LLMEmptyResponseError with specific user guidance
+            from penguin.utils.errors import LLMEmptyResponseError
+            
+            if isinstance(e, LLMEmptyResponseError):
+                error_msg = str(e)
+                logger.error(f"LLM empty response error: {error_msg}")
+                
+                # Provide helpful guidance to the user
+                user_message = (
+                    f"🐧 Penguin: I encountered an error - the model returned empty responses.\n\n"
+                    f"This may indicate:\n"
+                    f"  1. Context window exceeded (conversation too long)\n"
+                    f"  2. API quota or rate limit reached\n"
+                    f"  3. Model refusing to respond\n\n"
+                    f"Try starting a new conversation or checking your API status."
+                )
+                
+                await self._emit_event({
+                    "type": "error",
+                    "source": "llm_empty_response",
+                    "message": user_message,
+                    "details": {"error": error_msg}
+                })
+                
+                return {
+                    "status": "error",
+                    "message": user_message,
+                    "completion_type": "llm_empty_response_error"
+                }
+            
+            # Handle other execution errors
             error_msg = f"Error executing task: {str(e)}"
             logger.error(error_msg)
             logger.exception("Traceback for task execution error:")
@@ -915,4 +993,4 @@ class RunMode:
             return "continuous"
             
         # Standard task completion
-        return "task" 
\ No newline at end of file
+        return "task" 
