@@ -6,8 +6,10 @@ import httpx
 import asyncio
 from pathlib import Path
 import platform
+import shutil
 import subprocess
 from typing import Dict, Any, List, Optional, Tuple
+import shutil
 from rich.console import Console
 
 # Create rich console for enhanced output
@@ -435,7 +437,11 @@ async def run_setup_wizard() -> Dict[str, Any]:
             style=STYLE
         ).ask_async()
         
-        console.print("[green]✓[/green] API key saved securely")
+        # Persist key to ~/.config/penguin/.env and export it
+        if _persist_api_key(actual_provider, api_key):
+            console.print("[green]✓[/green] API key saved to ~/.config/penguin/.env and exported for this session")
+        else:
+            console.print("[yellow]⚠️ Could not persist API key automatically. It will be shown in next steps.[/yellow]")
     else:
         console.print("[yellow]ℹ️ No API key provided. You'll need to set this up later.[/yellow]")
     
@@ -599,10 +605,11 @@ async def run_setup_wizard() -> Dict[str, Any]:
     ).ask_async():
         # Show saving indicator
         with console.status("[bold cyan]Saving configuration...[/bold cyan]", spinner="dots"):
-            save_success = save_config(config)
+            saved_config_path = save_config(config)
         
-        if save_success:
+        if saved_config_path:
             console.print("[bold green]✓ Configuration saved successfully![/bold green]")
+            console.print(f"[dim]Saved to:[/dim] {saved_config_path}")
             
             # Mark setup as complete
             mark_setup_complete()
@@ -635,11 +642,22 @@ async def run_setup_wizard() -> Dict[str, Any]:
                     default=False,
                     style=STYLE
                 ).ask_async():
-                    config_path = get_config_path()
-                    if open_in_default_editor(config_path):
-                        console.print(f"[green]✓ Opened config file:[/green] {config_path}")
+                    cfg_path = saved_config_path
+                    # In Codespaces/containers, prefer $EDITOR if available, else fallback
+                    editor = os.environ.get("EDITOR")
+                    opened = False
+                    if editor and shutil.which(editor.split()[0]):
+                        try:
+                            subprocess.run([*editor.split(), str(cfg_path)], check=True)
+                            opened = True
+                        except Exception:
+                            opened = False
+                    if not opened:
+                        opened = open_in_default_editor(cfg_path)
+                    if opened:
+                        console.print(f"[green]✓ Opened config file:[/green] {cfg_path}")
                     else:
-                        console.print(f"[yellow]⚠️ Could not open config file. It's located at:[/yellow] {config_path}")
+                        console.print(f"[yellow]⚠️ Could not open config file. It's located at:[/yellow] {cfg_path}")
         else:
             console.print("[bold red]Failed to save configuration. Please check permissions and try again.[/bold red]")
     else:
@@ -655,64 +673,75 @@ async def run_setup_wizard() -> Dict[str, Any]:
     
     return config
 
-def get_config_path() -> Path:
-    """Return the path to the config file"""
-    # First check for explicit override
-    if os.environ.get("PENGUIN_CONFIG_PATH"):
-        return Path(os.environ.get("PENGUIN_CONFIG_PATH"))
-    
-    # For user configuration, we want to use a user-specific location
-    # that's separate from the package's default config
-    
-    # Check if we're in development mode (running from source)
-    try:
-        from penguin.config import get_project_root
-        project_root = get_project_root()
-        
-        # If PROJECT_ROOT points to a development directory (not .penguin), 
-        # use it for config (development mode)
-        if not str(project_root).endswith('.penguin'):
-            dev_config_path = project_root / "penguin" / "config.yml"
-            return dev_config_path
-    except Exception:
-        pass
-    
-    # For installed/production use, store user config in user's config directory
-    # Following XDG Base Directory specification on Linux/macOS
+def _user_config_dir() -> Path:
+    """Return cross-platform user config directory for Penguin."""
     if os.name == 'posix':  # Linux/macOS
-        config_base = Path(os.environ.get('XDG_CONFIG_HOME', Path.home() / '.config'))
-        user_config_path = config_base / "penguin" / "config.yml"
-    else:  # Windows
-        config_base = Path(os.environ.get('APPDATA', Path.home() / 'AppData' / 'Roaming'))
-        user_config_path = config_base / "penguin" / "config.yml"
-    
-    return user_config_path
+        return Path(os.environ.get('XDG_CONFIG_HOME', Path.home() / '.config')) / 'penguin'
+    return Path(os.environ.get('APPDATA', Path.home() / 'AppData' / 'Roaming')) / 'penguin'
 
-def save_config(config: Dict[str, Any]) -> bool:
+def get_config_path() -> Path:
+    """Return the path to the user config file (env override wins)."""
+    # 1) Explicit override
+    if os.environ.get("PENGUIN_CONFIG_PATH"):
+        return Path(os.environ["PENGUIN_CONFIG_PATH"]).expanduser()
+    # 2) User config directory (do not write into repo path)
+    return _user_config_dir() / "config.yml"
+
+def _persist_api_key(provider: str, api_key: str) -> bool:
+    """Write provider API key to ~/.config/penguin/.env and export in-process.
+
+    Returns True on success, False on failure.
     """
-    Save the config to a YAML file
-    
+    try:
+        env_dir = _user_config_dir()
+        env_dir.mkdir(parents=True, exist_ok=True)
+        env_path = env_dir / ".env"
+        # Read existing lines (if any)
+        existing: list[str] = []
+        if env_path.exists():
+            existing = env_path.read_text(encoding="utf-8").splitlines()
+        key = f"{provider.upper()}_API_KEY"
+        kv = f"{key}={api_key}"
+        replaced = False
+        for i, line in enumerate(existing):
+            if line.startswith(f"{key}="):
+                existing[i] = kv
+                replaced = True
+                break
+        if not replaced:
+            existing.append(kv)
+        env_path.write_text("\n".join(existing) + "\n", encoding="utf-8")
+        # Export for current process so the rest of the run sees it
+        os.environ[key] = api_key
+        return True
+    except Exception:
+        return False
+
+def save_config(config: Dict[str, Any]) -> Optional[Path]:
+    """
+    Save the config to a YAML file and return the resolved path.
+
     Returns:
-        bool: True if save was successful, False otherwise
+        Path on success; None on failure.
     """
     config_path = get_config_path()
-    
+
     try:
         # Make sure parent directory exists
         config_path.parent.mkdir(parents=True, exist_ok=True)
-        
+
         # Write config with nice formatting
         with open(config_path, 'w') as f:
             yaml.dump(config, f, sort_keys=False, default_flow_style=False)
-        
-        return True
+
+        return config_path
     except PermissionError:
         console.print(f"[bold red]⚠️ Permission denied:[/bold red] Cannot write to {config_path}")
         console.print("[yellow]Try running with administrator/sudo privileges or choose a different location.[/yellow]")
-        return False
+        return None
     except Exception as e:
         console.print(f"[bold red]⚠️ Error saving configuration:[/bold red] {str(e)}")
-        return False
+        return None
 
 def open_in_default_editor(file_path: Path) -> bool:
     """
@@ -731,10 +760,22 @@ def open_in_default_editor(file_path: Path) -> bool:
             elif platform.system() == 'Darwin':  # macOS
                 subprocess.run(['open', file_path], check=True)
             else:  # Linux
+                # In headless environments (e.g., Codespaces) xdg-open may not exist
+                if shutil.which('xdg-open') is None:
+                    console.print(
+                        f"[yellow]⚠️ GUI launcher not available (xdg-open not found). File saved at:[/yellow] {file_path}"
+                    )
+                    return False
                 subprocess.run(['xdg-open', file_path], check=True)
             return True
-        except FileNotFoundError:
-                console.print(f"[bold red]⚠️ File not found:[/bold red] {file_path}")
+        except FileNotFoundError as e:
+                # Differentiate missing editor launcher vs missing file
+                if file_path.exists():
+                    console.print(
+                        f"[yellow]⚠️ Could not launch editor ({e}). File saved at:[/yellow] {file_path}"
+                    )
+                else:
+                    console.print(f"[bold red]⚠️ File not found:[/bold red] {file_path}")
                 return False
         except PermissionError:
                 console.print(f"[bold red]⚠️ Permission denied:[/bold red] Cannot open {file_path}")
@@ -972,4 +1013,4 @@ def run_setup_wizard_sync() -> Dict[str, Any]:
         return {"error": "Setup interrupted"}
     except Exception as e:
         console.print(f"[red]Setup wizard error: {e}[/red]")
-        return {"error": str(e)} 
\ No newline at end of file
+        return {"error": str(e)} 
