@@ -10,6 +10,28 @@ This module provides tools to:
 4. Ensure context windows don't exceed model limits
 """
 
+# Future ideas:
+# - Make the budgets dynamic based on the content of the messages. Maybe?
+# - Don't hardcode the categories, make them configurable. Which means here it deals with the budgets of categories based on the priority they're infered from the config.
+
+# for multi/sub-agents, each will have their own context window manager that sits in that agent's shared context window manager
+# they will also be able to inherit some settings from the parent context window manager
+# these can't be too big, but they can't be too small either.
+# this is one of the reasons i wanted to split out conversation modules (where the conversation can be processed like a queue)
+# if its in that, then its a different conversation as far as the context window manager is concerned
+# in theory, this could extend to tool calls themselves, that could then have their own context window manager
+# yeah, an agent calling a tool that calls a tool that calls a tool etc
+
+"""
+Possibly in terms of Cognitive Architectures, Context Windows could be representated as States/Objects 
+that can be accessed by more than just one Agent. 
+
+Meaning you could have sub-agents that represent functions similar to what a Human brain could do 
+in terms of short/long term memory.
+"""
+
+# Also worth having a features / roadmap / competitors analysis.md files, for Penguin and what others are doing.
+
 # TODO: 
 # - Add a function to get the total token usage for a session
 # - Add a function to get the token usage for a specific category
@@ -25,6 +47,7 @@ This module provides tools to:
 
 import logging
 from dataclasses import dataclass, field
+from datetime import datetime
 from typing import Dict, List, Optional, Union, Any, Callable, Tuple
 
 from penguin.system.state import Message, MessageCategory, Session
@@ -40,13 +63,85 @@ class TokenBudget:
     max_tokens: int
     # Current tokens used by this category
     current_tokens: int = 0
-    
+
     def __post_init__(self):
         # Ensure min_tokens <= max_tokens
         if self.min_tokens > self.max_tokens:
             logger.warning(f"min_tokens ({self.min_tokens}) > max_tokens ({self.max_tokens}). Setting min_tokens = max_tokens")
             self.min_tokens = self.max_tokens
 
+
+@dataclass
+class TruncationEvent:
+    """Record of a context window truncation event"""
+    category: MessageCategory
+    messages_removed: int
+    tokens_freed: int
+    timestamp: str
+    total_messages_before: int
+    total_messages_after: int
+
+
+class TruncationTracker:
+    """Tracks context window truncation events for UI display"""
+
+    def __init__(self):
+        self.events: List[TruncationEvent] = []
+        self.per_category: Dict[MessageCategory, List[TruncationEvent]] = {
+            cat: [] for cat in MessageCategory
+        }
+        self.session_total_truncations = 0
+        self.session_total_messages_removed = 0
+        self.session_total_tokens_freed = 0
+
+    def record_truncation(
+        self,
+        category: MessageCategory,
+        messages_removed: int,
+        tokens_freed: int,
+        total_messages_before: int,
+        total_messages_after: int
+    ) -> TruncationEvent:
+        """Record a truncation event and return it"""
+        event = TruncationEvent(
+            category=category,
+            messages_removed=messages_removed,
+            tokens_freed=tokens_freed,
+            timestamp=datetime.now().isoformat(),
+            total_messages_before=total_messages_before,
+            total_messages_after=total_messages_after
+        )
+
+        self.events.append(event)
+        self.per_category[category].append(event)
+        self.session_total_truncations += 1
+        self.session_total_messages_removed += messages_removed
+        self.session_total_tokens_freed += tokens_freed
+
+        logger.info(
+            f"Truncation recorded: {category.name} - "
+            f"{messages_removed} messages, {tokens_freed} tokens freed"
+        )
+
+        return event
+
+    def get_category_truncations(self, category: MessageCategory) -> int:
+        """Get total number of messages truncated for a category"""
+        return sum(e.messages_removed for e in self.per_category[category])
+
+    def get_recent_events(self, limit: int = 10) -> List[TruncationEvent]:
+        """Get most recent truncation events"""
+        return self.events[-limit:] if self.events else []
+
+    def reset(self):
+        """Reset all truncation tracking"""
+        self.events.clear()
+        for cat_list in self.per_category.values():
+            cat_list.clear()
+        self.session_total_truncations = 0
+        self.session_total_messages_removed = 0
+        self.session_total_tokens_freed = 0
+
 class ContextWindowManager:
     """Manages token budgeting and content trimming for conversation context"""
     
@@ -54,24 +149,42 @@ class ContextWindowManager:
         self,
         model_config = None,
         token_counter: Optional[Callable[[Any], int]] = None,
-        api_client=None
+        api_client=None,
+        config_obj: Optional[Any] = None,
     ):
         """
         Initialize the context window manager.
-        
+
         Args:
             model_config: Optional model configuration with max_tokens
             token_counter: Function to count tokens for content
             api_client: API client for token counting
         """
-        # Get max_tokens from model_config if available
+        # Initialize truncation tracker
+        self.truncation_tracker = TruncationTracker()
+
+        # Get max_tokens from model_config / config context_window when available
         self.max_tokens = 150000  # Default fallback
         
         if model_config and hasattr(model_config, 'max_tokens') and model_config.max_tokens:
             self.max_tokens = model_config.max_tokens
             logger.info(f"Using model's max_tokens: {self.max_tokens}")
         
-        # Try to load from config.yml if available
+        # Prefer values coming from the live Config instance when provided
+        if config_obj is not None:
+            try:
+                cw_from_config = None
+                # Prefer a dedicated max_history_tokens (context capacity) when present
+                if hasattr(config_obj, 'model_config') and hasattr(config_obj.model_config, 'max_history_tokens'):
+                    cw_from_config = config_obj.model_config.max_history_tokens
+                # Upgrade only if larger
+                if cw_from_config and cw_from_config > self.max_tokens:
+                    self.max_tokens = cw_from_config
+                    logger.info(f"Using live Config max_history_tokens: {self.max_tokens}")
+            except Exception as e:
+                logger.warning(f"Failed to read context window from live Config: {e}")
+
+        # Try to load from module-level config as a last resort (legacy path)
         try:
             from penguin.config import config
             if 'model_configs' in config:
@@ -81,6 +194,17 @@ class ContextWindowManager:
                     if config_max_tokens:
                         self.max_tokens = config_max_tokens
                         logger.info(f"Using config.yml max_tokens for {model_name}: {self.max_tokens}")
+            # Prefer global model.context_window if present. Do not downgrade the
+            # window if model_config already provides a larger, authoritative value
+            # (e.g., fetched from model specs). Use the larger of the two.
+            if 'model' in config and isinstance(config['model'], dict):
+                cw = config['model'].get('context_window')
+                if cw:
+                    if self.max_tokens and cw > self.max_tokens:
+                        self.max_tokens = cw
+                        logger.info(f"Using global context_window from config.yml (upgraded): {self.max_tokens}")
+                    else:
+                        logger.info(f"Retaining model's max_tokens ({self.max_tokens}); global context_window={cw} not larger")
         except (ImportError, AttributeError) as e:
             logger.warning(f"Could not load config.yml for max_tokens: {e}")
 
@@ -120,6 +244,7 @@ class ContextWindowManager:
                 if isinstance(item, dict) and item.get("type") in ["image", "image_url"]:
                     # Much more realistic image token estimates - Claude models use ~4000 tokens per image
                     total += 4000  # Higher default for safety
+                    # TODO: Make this dynamic based on the model used
                 elif isinstance(item, dict) and "text" in item:
                     total += len(item["text"]) // 4
                 else:
@@ -344,7 +469,7 @@ class ContextWindowManager:
                 msgs_to_keep = []
                 for msg in category_msgs:
                     msg_tokens = msg.tokens or self.token_counter(msg.content)
-                    
+
                     if tokens_removed < category_excess:
                         # Remove this message (oldest first)
                         tokens_removed += msg_tokens
@@ -355,17 +480,33 @@ class ContextWindowManager:
                     else:
                         # Keep this message
                         msgs_to_keep.append(msg)
-                
+
                 # Replace the category's messages with trimmed version
                 categorized[category] = msgs_to_keep
-                
+
                 # Update statistics for next category
                 stats["per_category"][category] = category_tokens - tokens_removed
                 stats["total_tokens"] -= tokens_removed
-                
+
+                # Calculate messages removed count
+                messages_removed_count = len(category_msgs) - len(msgs_to_keep)
+
                 # Log the trimming operation
-                logger.debug(f"Trimmed {category.name}: removed {tokens_removed} tokens " + 
-                            f"({len(category_msgs) - len(msgs_to_keep)} messages)")
+                logger.debug(f"Trimmed {category.name}: removed {tokens_removed} tokens " +
+                            f"({messages_removed_count} messages)")
+
+                # Record truncation event if messages were actually removed
+                if messages_removed_count > 0:
+                    total_messages_before = len(session.messages)
+                    total_messages_after = total_messages_before - messages_removed_count
+
+                    self.truncation_tracker.record_truncation(
+                        category=category,
+                        messages_removed=messages_removed_count,
+                        tokens_freed=tokens_removed,
+                        total_messages_before=total_messages_before,
+                        total_messages_after=total_messages_after
+                    )
                 
                 # Check if we've trimmed enough overall
                 if tokens_to_trim <= 0:
@@ -758,4 +899,168 @@ class ContextWindowManager:
     def get_usage(self, category: MessageCategory) -> int:
         """Get current token usage for a specific category."""
         budget = self._budgets.get(category)
-        return budget.current_tokens if budget else 0 
\ No newline at end of file
+        return budget.current_tokens if budget else 0
+    
+    # =================================================================
+    # Phase 2 Extensions: Dynamic Reallocation and Context Contributors
+    # =================================================================
+    
+    def get_available_tokens(self, category: MessageCategory) -> int:
+        """Get available tokens for a category (max - current)"""
+        budget = self._budgets.get(category)
+        if not budget:
+            return 0
+        return max(0, budget.max_tokens - budget.current_tokens)
+    
+    def borrow_tokens(self, from_category: MessageCategory, to_category: MessageCategory, amount: int) -> bool:
+        """
+        Borrow unused tokens from one category to another.
+        
+        Args:
+            from_category: Category to borrow tokens from
+            to_category: Category to lend tokens to
+            amount: Number of tokens to borrow
+            
+        Returns:
+            True if borrowing was successful
+        """
+        from_budget = self._budgets.get(from_category)
+        to_budget = self._budgets.get(to_category)
+        
+        if not from_budget or not to_budget:
+            return False
+        
+        # Check if from_category has enough unused tokens
+        available_from = from_budget.max_tokens - from_budget.current_tokens
+        if available_from < amount:
+            return False
+        
+        # Perform the borrowing
+        from_budget.max_tokens -= amount
+        to_budget.max_tokens += amount
+        
+        logger.debug(f"Borrowed {amount} tokens from {from_category.name} to {to_category.name}")
+        return True
+    
+    def auto_rebalance_budgets(self) -> Dict[str, int]:
+        """
+        Automatically rebalance budgets based on current usage patterns.
+        Working files can borrow from DIALOG when needed.
+        
+        Returns:
+            Dict showing token movements
+        """
+        movements = {}
+        
+        # Priority for borrowing: CONTEXT (working files) can borrow from DIALOG
+        context_budget = self._budgets.get(MessageCategory.CONTEXT)
+        dialog_budget = self._budgets.get(MessageCategory.DIALOG)
+        
+        if context_budget and dialog_budget:
+            # If CONTEXT is over budget and DIALOG has unused tokens
+            context_over = context_budget.current_tokens - context_budget.max_tokens
+            dialog_available = dialog_budget.max_tokens - dialog_budget.current_tokens
+            
+            if context_over > 0 and dialog_available > 0:
+                # Borrow up to what's needed or available
+                borrow_amount = min(context_over, dialog_available, dialog_available // 2)  # Max 50% of available
+                
+                if self.borrow_tokens(MessageCategory.DIALOG, MessageCategory.CONTEXT, borrow_amount):
+                    movements[f"DIALOG -> CONTEXT"] = borrow_amount
+        
+        return movements
+    
+    def load_project_instructions(self, workspace_root: str = ".") -> Tuple[str, Dict[str, Any]]:
+        """
+        Load project instruction files (PENGUIN.md, AGENTS.md, README.md) with simple autoloading.
+        Simplified version without complex contributor system.
+        
+        Args:
+            workspace_root: Root directory of the workspace
+            
+        Returns:
+            Tuple of (project_content, debug_info)
+        """
+        from pathlib import Path
+        
+        workspace = Path(workspace_root).resolve()
+        content_parts = []
+        loaded_files = []
+        
+        # Priority order: PENGUIN.md > AGENTS.md > README.md  
+        doc_files = [
+            ("PENGUIN.md", "Project Instructions", 2400),  # 600 tokens worth
+            ("AGENTS.md", "Agent Specifications", 2000),   # 500 tokens worth
+            ("README.md", "Project Overview", 1200)        # 300 tokens worth  
+        ]
+        
+        for filename, section_title, max_chars in doc_files:
+            doc_path = workspace / filename
+            if doc_path.exists():
+                try:
+                    with open(doc_path, 'r', encoding='utf-8') as f:
+                        content = f.read()
+                        # Limit content to prevent context overflow
+                        if len(content) > max_chars:
+                            content = content[:max_chars] + "\n... (truncated)"
+                        content_parts.append(f"## {section_title} ({filename})\n{content}")
+                        loaded_files.append(filename)
+                        
+                        # If we found PENGUIN.md or AGENTS.md, don't fall back to README
+                        if filename in ["PENGUIN.md", "AGENTS.md"]:
+                            break
+                            
+                except Exception as e:
+                    logger.warning(f"Failed to read {filename}: {e}")
+        
+        assembled_content = "\n\n".join(content_parts) if content_parts else ""
+        token_count = self.token_counter(assembled_content) if assembled_content else 0
+
+        # Compute truncation flag correctly by re-evaluating based on the
+        # stored parts lengths versus their caps. Track truncation during load
+        # to avoid referencing out-of-scope variables.
+        truncated = False
+        try:
+            # If any chosen file's content was longer than its max_chars, we appended
+            # the marker; use that as a reliable signal.
+            truncated = any(part.endswith("... (truncated)") or "(truncated)" in part for part in content_parts)
+        except Exception:
+            truncated = False
+        
+        debug_info = {
+            "loaded_files": loaded_files,
+            "total_tokens": token_count,
+            "files_checked": [f[0] for f in doc_files],
+            "truncated": truncated,
+        }
+        
+        return assembled_content, debug_info
+    
+    def get_allocation_report(self) -> Dict[str, Any]:
+        """
+        Generate a detailed allocation report for debugging.
+        Shows how tokens are allocated across categories and contributors.
+        """
+        report = {
+            "total_budget": self.max_tokens,
+            "categories": {},
+            "utilization": {}
+        }
+        
+        total_used = 0
+        for category, budget in self._budgets.items():
+            category_info = {
+                "min_tokens": budget.min_tokens,
+                "max_tokens": budget.max_tokens,
+                "current_tokens": budget.current_tokens,
+                "available": budget.max_tokens - budget.current_tokens,
+                "utilization_pct": (budget.current_tokens / budget.max_tokens * 100) if budget.max_tokens else 0
+            }
+            report["categories"][category.name] = category_info
+            total_used += budget.current_tokens
+        
+        report["utilization"]["total_used"] = total_used
+        report["utilization"]["total_available"] = self.max_tokens - total_used
+        report["utilization"]["overall_pct"] = (total_used / self.max_tokens * 100) if self.max_tokens else 0
+        
+        return report 
