@@ -13,6 +13,11 @@ from typing import Any, Dict, List, Optional, Tuple, Union
 from penguin.config import CONVERSATION_CONFIG
 from penguin.system.state import Message, MessageCategory, Session
 from penguin.utils.diagnostics import diagnostics
+try:
+    from penguin.system.message_bus import MessageBus, ProtocolMessage
+except Exception:  # pragma: no cover - optional import to avoid cycles in minimal envs
+    MessageBus = None  # type: ignore
+    ProtocolMessage = None  # type: ignore
 
 # Optional - can be replaced with approximation method for multiple providers
 try:
@@ -70,24 +75,35 @@ class ConversationSystem:
         self.system_prompt_sent = False
 
     def add_message(
-        self, 
-        role: str, 
+        self,
+        role: str,
         content: Any,
-        category: MessageCategory = None, 
-        metadata: Optional[Dict[str, Any]] = None
+        category: MessageCategory = None,
+        metadata: Optional[Dict[str, Any]] = None,
+        *,
+        agent_id: Optional[str] = None,
+        recipient_id: Optional[str] = None,
+        message_type: str = "message",
     ) -> Message:
         """
         Add a message to the current session.
-        
+
         Args:
             role: Message role (system, user, assistant)
             content: Message content (string, list, or dict)
             category: Message category for token budgeting
             metadata: Optional metadata for the message
-            
+
         Returns:
             The created Message object
         """
+        # Debug: log when assistant messages are added
+        if role == "assistant":
+            import traceback
+            print(f"[DEBUG] add_message() called for assistant. content_length={len(str(content))}, has_reasoning={metadata.get('has_reasoning') if metadata else False}", flush=True)
+            print(f"[DEBUG] Stack trace:", flush=True)
+            for line in traceback.format_stack()[-5:-1]:
+                print(line.strip(), flush=True)
         # Set default category based on role if not specified
         if category is None:
             if role == "system":
@@ -101,6 +117,13 @@ class ConversationSystem:
             else:
                 category = MessageCategory.DIALOG
         
+        # Resolve agent_id default from session metadata if not provided
+        if agent_id is None:
+            try:
+                agent_id = self.session.metadata.get("agent_id")
+            except Exception:
+                agent_id = None
+
         # Create the message
         message = Message(
             role=role,
@@ -109,11 +132,44 @@ class ConversationSystem:
             id=f"msg_{uuid.uuid4().hex[:8]}",
             timestamp=datetime.now().isoformat(),
             metadata=metadata or {},
+            tokens=0,
+            agent_id=agent_id,
+            recipient_id=recipient_id,
+            message_type=message_type,
         )
         
         # Add to current session
         self.session.messages.append(message)
         self._modified = True
+
+        # Phase 3: publish protocol message to MessageBus (best-effort)
+        try:
+            if MessageBus and ProtocolMessage:
+                bus = MessageBus.get_instance()
+                pm = ProtocolMessage(
+                    sender=message.agent_id,
+                    recipient=None,
+                    content=message.content,
+                    message_type=message.message_type,
+                    metadata={
+                        **(message.metadata or {}),
+                        "category": message.category.name,
+                        "role": message.role,
+                    },
+                    session_id=self.session.id,
+                    message_id=message.id,
+                )
+                # Fire-and-forget, don't block
+                import asyncio as _asyncio
+                try:
+                    loop = _asyncio.get_event_loop()
+                    if loop.is_running():
+                        _asyncio.create_task(bus.send(pm))
+                except RuntimeError:
+                    # No running loop; skip bus delivery in sync contexts
+                    pass
+        except Exception:
+            pass
         
         # NEW: Auto-checkpoint integration
         if self.checkpoint_manager and self.checkpoint_manager.should_checkpoint(message):
@@ -182,7 +238,8 @@ class ConversationSystem:
                 "system", 
                 self.system_prompt, 
                 MessageCategory.SYSTEM,
-                {"type": "system_prompt"}
+                {"type": "system_prompt"},
+                message_type="status",
             )
             self.system_prompt_sent = True
 
@@ -264,7 +321,8 @@ class ConversationSystem:
                 "tool_call_id": tool_call_id,
                 "action_type": action_type,
                 "status": status
-            }
+            },
+            message_type="action",
         )
 
     def add_context(
@@ -309,7 +367,8 @@ class ConversationSystem:
             "system",
             content,
             MessageCategory.SYSTEM_OUTPUT,
-            {"type": "iteration_marker", "iteration": iteration}
+            {"type": "iteration_marker", "iteration": iteration},
+            message_type="status",
         )
 
     def get_history(self) -> List[Dict[str, Any]]:
@@ -416,30 +475,36 @@ class ConversationSystem:
         
     def load(self, session_id: str) -> bool:
         """
-        Load a session by ID via session manager.
-        
+        Load a session by ID via session manager, or create a new one if it doesn't exist.
+
         Args:
             session_id: ID of the session to load
-            
+
         Returns:
-            True if successful, False otherwise
+            True if successful (loaded or created), False on error
         """
         if not self.session_manager:
             logger.error("Cannot load: No session manager available")
             return False
-            
+
         loaded_session = self.session_manager.load_session(session_id)
         if loaded_session:
             self.session = loaded_session
             self._modified = False
             # Update sent status
             self.system_prompt_sent = any(
-                msg.category == MessageCategory.SYSTEM 
+                msg.category == MessageCategory.SYSTEM
                 for msg in self.session.messages
             )
+            logger.debug(f"Loaded existing session: {session_id}")
+            return True
+        else:
+            # Session doesn't exist - create a new one with this ID
+            logger.info(f"Session {session_id} not found, creating new session")
+            self.session = Session(id=session_id)
+            self._modified = True  # Mark as modified so it gets saved
+            self.system_prompt_sent = False
             return True
-            
-        return False
         
     def load_context_file(self, file_path: str) -> bool:
         """
@@ -499,4 +564,4 @@ class ConversationSystem:
         self._modified = True
         
         if self.session_manager:
-            self.session_manager.current_session = self.session
\ No newline at end of file
+            self.session_manager.current_session = self.session
