@@ -6,12 +6,13 @@
 # CodeAct Github: https://github.com/xingyaoww/code-act
 
 import asyncio
+import json
 import logging
 import re
 from datetime import datetime
 from enum import Enum
 from html import unescape
-from typing import List, Dict, Any
+from typing import List, Dict, Any, Optional, Callable, Awaitable
 import base64
 from penguin.local_task.manager import ProjectManager
 from penguin.tools import ToolManager
@@ -44,6 +45,7 @@ class ActionType(Enum):
     ENHANCED_READ = "enhanced_read"
     ENHANCED_WRITE = "enhanced_write"
     APPLY_DIFF = "apply_diff"
+    MULTIEDIT = "multiedit"
     EDIT_WITH_PATTERN = "edit_with_pattern"
     # TASK_CREATE = "task_create"
     # TASK_UPDATE = "task_update"
@@ -85,6 +87,7 @@ class ActionType(Enum):
     DEPENDENCY_DISPLAY = "dependency_display"
     ANALYZE_CODEBASE = "analyze_codebase"
     REINDEX_WORKSPACE = "reindex_workspace"
+    SEND_MESSAGE = "send_message"
     # Browser actions
     BROWSER_NAVIGATE = "browser_navigate"
     BROWSER_INTERACT = "browser_interact"
@@ -93,9 +96,16 @@ class ActionType(Enum):
     PYDOLL_BROWSER_NAVIGATE = "pydoll_browser_navigate"
     PYDOLL_BROWSER_INTERACT = "pydoll_browser_interact"
     PYDOLL_BROWSER_SCREENSHOT = "pydoll_browser_screenshot"
+    PYDOLL_BROWSER_SCROLL = "pydoll_browser_scroll"
     # PyDoll debug toggle
     PYDOLL_DEBUG_TOGGLE = "pydoll_debug_toggle"
     
+    # Sub-agent tools (agents-as-tools)
+    SPAWN_SUB_AGENT = "spawn_sub_agent"
+    STOP_SUB_AGENT = "stop_sub_agent"
+    RESUME_SUB_AGENT = "resume_sub_agent"
+    DELEGATE = "delegate"
+    
     # Repository management actions
     GET_REPOSITORY_STATUS = "get_repository_status"
     CREATE_AND_SWITCH_BRANCH = "create_and_switch_branch"
@@ -168,16 +178,71 @@ def parse_action(content: str) -> List[CodeActAction]:
 
 
 class ActionExecutor:
-    def __init__(self, tool_manager: ToolManager, task_manager: ProjectManager, conversation_system=None):
+    def __init__(
+        self,
+        tool_manager: ToolManager,
+        task_manager: ProjectManager,
+        conversation_system=None,
+        ui_event_callback: Optional[Callable[[str, Dict[str, Any]], Awaitable[None]]] = None,
+    ):
+        """Execute parsed actions and (optionally) emit UI events.
+        
+        Args:
+            tool_manager: The ToolManager instance
+            task_manager: The Project/Task manager
+            conversation_system: Conversation manager (optional)
+            ui_event_callback: Async callback to emit UI events (e.g. PenguinCore.emit_ui_event)
+        """
         self.tool_manager = tool_manager
         self.task_manager = task_manager
         self.process_manager = ProcessManager()
         self.current_process = None
         self.conversation_system = conversation_system
+        try:
+            logger.info(
+                "ActionExecutor initialized with ToolManager id=%s file_root=%s mode=%s",
+                hex(id(tool_manager)) if tool_manager else None,
+                getattr(tool_manager, '_file_root', None) if tool_manager else None,
+                getattr(tool_manager, 'file_root_mode', None) if tool_manager else None,
+            )
+        except Exception:
+            pass
+        self._ui_event_cb = ui_event_callback
         # No direct initialization of expensive tools, we'll use tool_manager's properties
 
     async def execute_action(self, action: CodeActAction) -> str:
+        """Execute an action and emit UI events if a callback is provided."""
         logger.debug(f"Attempting to execute action: {action.action_type.value}")
+        try:
+            logger.info(
+                "ActionExecutor executing %s using ToolManager id=%s file_root=%s mode=%s",
+                action.action_type.value,
+                hex(id(self.tool_manager)) if self.tool_manager else None,
+                getattr(self.tool_manager, '_file_root', None) if self.tool_manager else None,
+                getattr(self.tool_manager, 'file_root_mode', None) if self.tool_manager else None,
+            )
+        except Exception:
+            pass
+        import uuid
+        action_id = str(uuid.uuid4())[:8]
+        
+        # --------------------------------------------------
+        # Emit *start* UI event
+        # --------------------------------------------------
+        if self._ui_event_cb:
+            try:
+                logger.debug(f"UI-emit start action id={action_id} type={action.action_type.value}")
+                await self._ui_event_cb(
+                    "action",
+                    {
+                        "id": action_id,
+                        "type": action.action_type.value,
+                        "params": action.params,
+                    },
+                )
+            except Exception as e:
+                logger.debug(f"UI event emit failed (action start): {e}")
+        
         action_map = {
             # ActionType.READ: lambda params: self.tool_manager.execute_tool("read_file", {"path": params}),
             # ActionType.WRITE: self._write_file,
@@ -216,6 +281,7 @@ class ActionExecutor:
             ActionType.PROCESS_SEND: self._process_send,
             ActionType.PROCESS_EXIT: self._process_exit,
             ActionType.WORKSPACE_SEARCH: self._workspace_search,
+            ActionType.MULTIEDIT: self._multiedit,
             # Project management handlers
             ActionType.PROJECT_CREATE: self._project_create,
             ActionType.PROJECT_LIST: self._project_list,
@@ -232,6 +298,7 @@ class ActionExecutor:
             ActionType.DEPENDENCY_DISPLAY: self._dependency_display,
             ActionType.ANALYZE_CODEBASE: self._analyze_codebase,
             ActionType.REINDEX_WORKSPACE: self._reindex_workspace,
+            ActionType.SEND_MESSAGE: self._send_message,
             # Browser actions
             ActionType.BROWSER_NAVIGATE: self._browser_navigate,
             ActionType.BROWSER_INTERACT: self._browser_interact,
@@ -240,8 +307,14 @@ class ActionExecutor:
             ActionType.PYDOLL_BROWSER_NAVIGATE: self._pydoll_browser_navigate,
             ActionType.PYDOLL_BROWSER_INTERACT: self._pydoll_browser_interact,
             ActionType.PYDOLL_BROWSER_SCREENSHOT: self._pydoll_browser_screenshot,
+            ActionType.PYDOLL_BROWSER_SCROLL: self._pydoll_browser_scroll,
             # PyDoll debug toggle
             ActionType.PYDOLL_DEBUG_TOGGLE: self._pydoll_debug_toggle,
+            # Sub-agent tools
+            ActionType.SPAWN_SUB_AGENT: self._spawn_sub_agent,
+            ActionType.STOP_SUB_AGENT: self._stop_sub_agent,
+            ActionType.RESUME_SUB_AGENT: self._resume_sub_agent,
+            ActionType.DELEGATE: self._delegate,
             # Enhanced file operations
             ActionType.LIST_FILES_FILTERED: self._list_files_filtered,
             ActionType.FIND_FILES_ENHANCED: self._find_files_enhanced,
@@ -278,12 +351,43 @@ class ActionExecutor:
                     result = await result
 
             logger.info(f"Action {action.action_type.value} executed successfully")
+            # --------------------------------------------------
+            # Emit success UI event
+            # --------------------------------------------------
+            if self._ui_event_cb:
+                try:
+                    logger.debug(f"UI-emit success action_result id={action_id}")
+                    await self._ui_event_cb(
+                        "action_result",
+                        {
+                            "id": action_id,
+                            "status": "completed",
+                            "result": result if isinstance(result, str) else str(result),
+                            "action_name": action.action_type.value,
+                        },
+                    )
+                except Exception as e:
+                    logger.debug(f"UI event emit failed (action result): {e}")
             return result
         except Exception as e:
             error_message = (
                 f"Error executing action {action.action_type.value}: {str(e)}"
             )
             logger.error(error_message, exc_info=True)
+            if self._ui_event_cb:
+                try:
+                    logger.debug(f"UI-emit error action_result id={action_id}")
+                    await self._ui_event_cb(
+                        "action_result",
+                        {
+                            "id": action_id,
+                            "status": "error",
+                            "result": error_message,
+                            "action_name": action.action_type.value,
+                        },
+                    )
+                except Exception as ee:
+                    logger.debug(f"UI event emit failed (action error): {ee}")
             return error_message
 
     # def _write_file(self, params: str) -> str:
@@ -398,6 +502,284 @@ class ActionExecutor:
         self.current_process = None
         return result
 
+    async def _send_message(self, params: str) -> str:
+        try:
+            payload = json.loads(params) if params.strip().startswith("{") else {"content": params}
+        except json.JSONDecodeError as exc:
+            raise ValueError(f"send_message expects JSON payload: {exc}")
+
+        content = payload.get("content") or payload.get("message")
+        if not content:
+            raise ValueError("send_message requires 'content'")
+
+        channel = payload.get("channel")
+        message_type = payload.get("message_type", "message")
+        metadata = payload.get("metadata") or {}
+        sender = payload.get("sender")
+        raw_targets = payload.get("targets") or payload.get("target") or payload.get("recipient")
+
+        if raw_targets is None:
+            targets = ["human"]
+        elif isinstance(raw_targets, (list, tuple, set)):
+            targets = list(raw_targets)
+        else:
+            targets = [str(raw_targets)]
+
+        conversation = self.conversation_system
+        core = getattr(conversation, "core", None)
+
+        add_message_fn = None
+        save_fn = None
+        if callable(getattr(conversation, "add_message", None)):
+            add_message_fn = conversation.add_message
+            save_fn = getattr(conversation, "save", None)
+        elif hasattr(conversation, "conversation") and callable(
+            getattr(conversation.conversation, "add_message", None)
+        ):
+            add_message_fn = conversation.conversation.add_message
+            save_fn = getattr(conversation, "save", None)
+
+        if add_message_fn is None and core is None:
+            raise RuntimeError("Penguin core unavailable for send_message action")
+
+        results = []
+        for target in targets:
+            normalized = (target or "").strip()
+            if normalized in ("", "human", "user"):
+                delivered = False
+                if core is not None:
+                    try:
+                        delivered = await core.send_to_human(
+                            content,
+                            message_type=message_type,
+                            metadata=metadata,
+                            channel=channel,
+                        )
+                    except Exception as exc:  # pragma: no cover - defensive logging
+                        logger.warning("send_message core send_to_human failed: %s", exc)
+                        delivered = False
+                if not delivered and add_message_fn is not None:
+                    add_message_fn(
+                        role="assistant",
+                        content=content,
+                        category=MessageCategory.DIALOG,
+                        metadata={
+                            **metadata,
+                            "via": "send_message_fallback",
+                            "target": "human",
+                            "channel": channel,
+                            "message_type": message_type,
+                            "sender": sender,
+                        },
+                        message_type=message_type,
+                        agent_id=sender,
+                        recipient_id="human",
+                    )
+                    if callable(save_fn):
+                        try:
+                            save_fn()
+                        except Exception:  # pragma: no cover - best effort
+                            logger.debug("send_message fallback save failed", exc_info=True)
+                    results.append("human (logged)")
+                else:
+                    results.append("human")
+            else:
+                delivered = False
+                if core is not None:
+                    try:
+                        delivered = await core.route_message(
+                            normalized,
+                            content,
+                            message_type=message_type,
+                            metadata=metadata,
+                            agent_id=sender,
+                            channel=channel,
+                        )
+                    except Exception as exc:  # pragma: no cover - defensive logging
+                        logger.warning("send_message core route_message failed: %s", exc)
+                        delivered = False
+                if not delivered and add_message_fn is not None:
+                    add_message_fn(
+                        role="assistant",
+                        content=content,
+                        category=MessageCategory.DIALOG,
+                        metadata={
+                            **metadata,
+                            "via": "send_message_fallback",
+                            "target": normalized,
+                            "channel": channel,
+                            "message_type": message_type,
+                            "sender": sender,
+                        },
+                        message_type=message_type,
+                        agent_id=sender,
+                        recipient_id=normalized,
+                    )
+                    if callable(save_fn):
+                        try:
+                            save_fn()
+                        except Exception:  # pragma: no cover - best effort
+                            logger.debug("send_message fallback save failed", exc_info=True)
+                    results.append(f"{normalized} (logged)")
+                elif delivered:
+                    results.append(normalized)
+                else:
+                    results.append(f"{normalized} (failed)")
+
+        return f"Sent message to {', '.join(results)}"
+
+    # --------------------------------------------------
+    # Sub-agent tools (agents-as-tools)
+    # --------------------------------------------------
+
+    async def _spawn_sub_agent(self, params: str) -> str:
+        """Spawn a sub-agent; defaults to isolated session and context-window.
+
+        JSON body:
+          - id (required), parent (optional, default current), persona/system_prompt (optional)
+          - share_session (bool, default False), share_context_window (bool, default False)
+          - shared_cw_max_tokens (int, optional), model_* overrides (optional), default_tools (optional)
+          - initial_prompt (optional)
+        """
+        try:
+            payload = json.loads(params) if params.strip() else {}
+        except Exception as e:
+            return f"Invalid JSON for spawn_sub_agent: {e}"
+
+        agent_id = str(payload.get("id") or "").strip()
+        if not agent_id:
+            return "spawn_sub_agent requires 'id'"
+
+        conversation = self.conversation_system
+        core = getattr(conversation, "core", None)
+        if core is None:
+            return "Core unavailable for spawn_sub_agent"
+
+        parent_id = str(payload.get("parent") or getattr(conversation, "current_agent_id", None) or "default").strip()
+        share_session = bool(payload.get("share_session", False))
+        share_cw = bool(payload.get("share_context_window", False))
+        shared_cw_max_tokens = payload.get("shared_cw_max_tokens")
+        try:
+            shared_cw_max_tokens = int(shared_cw_max_tokens) if shared_cw_max_tokens is not None else None
+        except Exception:
+            shared_cw_max_tokens = None
+
+        kwargs: Dict[str, Any] = {}
+        for key in ("persona", "system_prompt", "model_config_id", "model_max_tokens", "default_tools"):
+            if key in payload:
+                kwargs[key] = payload[key]
+        if isinstance(payload.get("model_overrides"), dict):
+            kwargs["model_overrides"] = payload["model_overrides"]
+
+        try:
+            core.create_sub_agent(
+                agent_id,
+                parent_agent_id=parent_id,
+                share_session=share_session,
+                share_context_window=share_cw,
+                shared_cw_max_tokens=shared_cw_max_tokens,
+                **kwargs,
+            )
+        except Exception as e:
+            return f"Failed to spawn sub-agent '{agent_id}': {e}"
+
+        initial_prompt = payload.get("initial_prompt")
+        if initial_prompt:
+            try:
+                await core.send_to_agent(agent_id, initial_prompt)
+            except Exception:
+                pass
+
+        return f"Spawned sub-agent '{agent_id}' (parent='{parent_id}', share_session={share_session}, share_context_window={share_cw})"
+
+    async def _stop_sub_agent(self, params: str) -> str:
+        try:
+            payload = json.loads(params) if params.strip() else {}
+        except Exception as e:
+            return f"Invalid JSON for stop_sub_agent: {e}"
+        agent_id = str(payload.get("id") or "").strip()
+        if not agent_id:
+            return "stop_sub_agent requires 'id'"
+        conversation = self.conversation_system
+        core = getattr(conversation, "core", None)
+        if core is None:
+            return "Core unavailable for stop_sub_agent"
+        try:
+            if hasattr(core, "set_agent_paused"):
+                core.set_agent_paused(agent_id, True)
+            return f"Paused sub-agent '{agent_id}'"
+        except Exception as e:
+            return f"Failed to pause sub-agent '{agent_id}': {e}"
+
+    async def _resume_sub_agent(self, params: str) -> str:
+        try:
+            payload = json.loads(params) if params.strip() else {}
+        except Exception as e:
+            return f"Invalid JSON for resume_sub_agent: {e}"
+        agent_id = str(payload.get("id") or "").strip()
+        if not agent_id:
+            return "resume_sub_agent requires 'id'"
+        conversation = self.conversation_system
+        core = getattr(conversation, "core", None)
+        if core is None:
+            return "Core unavailable for resume_sub_agent"
+        try:
+            if hasattr(core, "set_agent_paused"):
+                core.set_agent_paused(agent_id, False)
+            return f"Resumed sub-agent '{agent_id}'"
+        except Exception as e:
+            return f"Failed to resume sub-agent '{agent_id}': {e}"
+
+    async def _delegate(self, params: str) -> str:
+        try:
+            payload = json.loads(params) if params.strip() else {}
+        except Exception as e:
+            return f"Invalid JSON for delegate: {e}"
+
+        child = str(payload.get("child") or "").strip()
+        content = payload.get("content")
+        if not child or content is None:
+            return "delegate requires 'child' and 'content'"
+
+        conversation = self.conversation_system
+        core = getattr(conversation, "core", None)
+        if core is None:
+            return "Core unavailable for delegate"
+
+        parent = str(payload.get("parent") or getattr(conversation, "current_agent_id", None) or "default").strip()
+        channel = payload.get("channel")
+        metadata = payload.get("metadata") or {}
+
+        # Record delegation event in both parent and child logs (best-effort)
+        try:
+            cm = getattr(core, "conversation_manager", None)
+            if cm and hasattr(cm, "log_delegation_event"):
+                import uuid as _uuid
+                delegation_id = _uuid.uuid4().hex[:8]
+                cm.log_delegation_event(
+                    delegation_id=delegation_id,
+                    parent_agent_id=parent,
+                    child_agent_id=child,
+                    event="request_sent",
+                    message=str(content)[:140],
+                    metadata={**metadata, **({"channel": channel} if channel else {})},
+                )
+                metadata = {**metadata, "delegation_id": delegation_id}
+        except Exception:
+            pass
+
+        try:
+            await core.send_to_agent(
+                child,
+                content,
+                message_type="message",
+                metadata=metadata,
+                channel=channel,
+            )
+            return f"Delegated to '{child}' from '{parent}'"
+        except Exception as e:
+            return f"Failed to delegate to '{child}': {e}"
+
     def _workspace_search(self, params: str) -> str:
         parts = params.split(":", 1)
         if len(parts) == 2:
@@ -886,6 +1268,52 @@ class ActionExecutor:
             logger.error(error_message, exc_info=True)
             return error_message
 
+    async def _pydoll_browser_scroll(self, params: str) -> str:
+        """Scroll the page or an element using PyDoll.
+        Formats:
+          - to:top|bottom
+          - page:down|up|end|home[:repeat]
+          - by:deltaY[:deltaX][:repeat]
+          - element:selector[:selector_type][:behavior]
+        Defaults: selector_type=css, behavior=auto, repeat=1
+        """
+        try:
+            from penguin.tools.pydoll_tools import PyDollBrowserScrollTool
+
+            parts = params.split(":") if params else []
+            if not parts:
+                return "Invalid scroll params"
+            mode = parts[0].strip().lower()
+
+            tool = PyDollBrowserScrollTool()
+
+            if mode == "to" and len(parts) >= 2:
+                to = parts[1].strip().lower()
+                return await tool.execute(mode="to", to=to)
+
+            if mode == "page":
+                direction = parts[1].strip().lower() if len(parts) >= 2 else "down"
+                repeat = int(parts[2]) if len(parts) >= 3 and parts[2].strip().isdigit() else 1
+                return await tool.execute(mode="page", to=direction, repeat=repeat)
+
+            if mode == "by":
+                dy = int(parts[1]) if len(parts) >= 2 and parts[1].strip().lstrip("-+").isdigit() else 800
+                dx = int(parts[2]) if len(parts) >= 3 and parts[2].strip().lstrip("-+").isdigit() else 0
+                repeat = int(parts[3]) if len(parts) >= 4 and parts[3].strip().isdigit() else 1
+                return await tool.execute(mode="by", delta_y=dy, delta_x=dx, repeat=repeat)
+
+            if mode == "element" and len(parts) >= 2:
+                selector = parts[1]
+                selector_type = parts[2].strip().lower() if len(parts) >= 3 and parts[2].strip() else "css"
+                behavior = parts[3].strip().lower() if len(parts) >= 4 and parts[3].strip() else "auto"
+                return await tool.execute(mode="element", selector=selector, selector_type=selector_type, behavior=behavior)
+
+            return "Invalid scroll command"
+        except Exception as e:
+            error_message = f"Error scrolling with PyDoll browser: {str(e)}"
+            logger.error(error_message)
+            return error_message
+
     async def _pydoll_debug_toggle(self, params: str) -> str:
         """Toggle PyDoll debug mode. Format: [on|off] or empty to toggle"""
         try:
@@ -1026,30 +1454,73 @@ class ActionExecutor:
 
     def _apply_diff(self, params: str) -> str:
         """Apply a diff to edit a file. Format: file_path:diff_content:backup"""
-        parts = params.split(":", 2)
-        if len(parts) < 2:
+        first_sep = params.find(":")
+        if first_sep == -1:
             return "Error: Need file path and diff content"
-        
-        file_path = parts[0].strip()
-        diff_content = parts[1]  # Don't strip - diff content has important formatting
-        backup = parts[2].strip().lower() == "true" if len(parts) > 2 else True
-        
+
+        file_path = params[:first_sep].strip()
+        remainder = params[first_sep + 1 :]
+        if not file_path or not remainder:
+            return "Error: Need file path and diff content"
+
+        backup = True
+        diff_content = remainder
+
+        # Optional backup flag can be appended as :true or :false with no newline.
+        if ":" in remainder:
+            diff_part, flag = remainder.rsplit(":", 1)
+            flag_stripped = flag.strip().lower()
+            if flag_stripped in {"true", "false"} and "\n" not in flag and "\r" not in flag:
+                backup = flag_stripped == "true"
+                diff_content = diff_part
+
         return self.tool_manager.execute_tool("apply_diff", {
             "file_path": file_path,
             "diff_content": diff_content,
             "backup": backup
         })
 
+    def _multiedit(self, params: str) -> str:
+        """Apply multi-file edits atomically. Content is the multiedit block or unified patch. Default dry-run."""
+        content = params
+        # Optional header directive: apply=true on the first line
+        do_apply = False
+        m = re.match(r"^apply\s*[:=]\s*(true|false)\s*\n", content, flags=re.IGNORECASE)
+        if m:
+            do_apply = m.group(1).lower() == 'true'
+            content = content[m.end():]
+        return self.tool_manager.execute_tool("multiedit_apply", {
+            "content": content,
+            "apply": do_apply,
+        })
+
     def _edit_with_pattern(self, params: str) -> str:
-        """Edit file with pattern replacement. Format: file_path:search_pattern:replacement:backup"""
-        parts = params.split(":", 3)
+        """Edit file with pattern replacement. Format: file_path:search_pattern:replacement:backup
+        
+        Note: Uses reverse split to handle colons in replacement text.
+        """
+        # Split from the right to handle optional backup flag first
+        parts = params.rsplit(":", 1)
+        backup = False
+        content_parts = parts[0]
+        
+        # Check if last part is the backup flag (true/false)
+        if len(parts) == 2 and parts[1].strip().lower() in ("true", "false"):
+            backup = parts[1].strip().lower() == "true"
+        else:
+            # No backup flag, treat entire string as content
+            content_parts = params
+            backup = True  # Default to backup
+        
+        # Now split the content part into file_path:search_pattern:replacement
+        # Split only on first 2 colons to preserve colons in replacement text
+        parts = content_parts.split(":", 2)
         if len(parts) < 3:
-            return "Error: Need file path, search pattern, and replacement"
+            return "Error: Need file_path:search_pattern:replacement format"
         
         file_path = parts[0].strip()
         search_pattern = parts[1]  # Don't strip - regex patterns may need whitespace
-        replacement = parts[2]      # Don't strip - replacement may need whitespace
-        backup = parts[3].strip().lower() == "true" if len(parts) > 3 else True
+        replacement = parts[2]      # Don't strip - replacement may need whitespace (and may contain colons!)
         
         return self.tool_manager.execute_tool("edit_with_pattern", {
             "file_path": file_path,
